VERSION = 'v1'

KONDUKTOR = 'KONDUKTOR'
KONDUKTOR_SERVICE_ACCOUNT_ID = KONDUKTOR + '-' + VERSION
KONDUKTOR_SERVICE_ACCOUNT_EMAIL_TEMPLATE = (
    '{account_id}@{project_id}.iam.gserviceaccount.com'
)
KONDUKTOR_SERVICE_ACCOUNT_CONFIG = {
    'displayName': f'KONDUKTOR Service Account ({VERSION})',
}

# Those roles will be always added.
# NOTE: `serviceAccountUser` allows the head node to create workers with
# a serviceAccount. `roleViewer` allows the head node to run bootstrap_gcp.
DEFAULT_SERVICE_ACCOUNT_ROLES = [
    'roles/storage.admin',
    'roles/iam.serviceAccountUser',
    'roles/iam.roleViewer',
]

# A list of permissions required to run Konduktor on GCP.
# Keep this in sync with https://docs.skypilot.co/en/latest/cloud-setup/cloud-permissions/gcp.html # noqa: E501
STORAGE_MINIMAL_PERMISSIONS = [
    'iam.roles.get',
    # We now skip the check for `iam.serviceAccounts.actAs` permission for
    # simplicity as it can be granted at the service-account level.
    # Check: sky.provision.gcp.config::_is_permission_satisfied
    # 'iam.serviceAccounts.actAs',
    'iam.serviceAccounts.get',
    'serviceusage.services.enable',
    'serviceusage.services.list',
    'serviceusage.services.use',
    'storage.buckets.create',
    'storage.buckets.get',
    'storage.buckets.delete',
    'storage.objects.create',
    'storage.objects.delete',
    'storage.objects.update',
    'storage.objects.get',
    'storage.objects.list',
    'resourcemanager.projects.get',
]



"""Data sync between workstation <--> blob (s3, gcs, etc.) <--> worker pods"""

from konduktor.data.storage import Storage, StorageMode, StoreType

__all__ = [
    'Storage',
    'StorageMode',
    'StoreType',
]



# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: skypilot: https://github.com/skypilot-org/skypilot
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lazy import for modules to avoid import error when not used."""

import functools
import importlib
import threading
from typing import Any, Callable, Optional, Tuple


class LazyImport:
    """Lazy importer for heavy modules or cloud modules only when enabled.

    We use this for pandas and networkx, as they can be time-consuming to import
    (0.1-0.2 seconds). With this class, we can avoid the unnecessary import time
    when the module is not used (e.g., `networkx` should not be imported for
    `sky status and `pandas` should not be imported for `sky exec`).

    We also use this for cloud adaptors, because we do not want to import the
    cloud dependencies when it is not enabled.
    """

    def __init__(
        self,
        module_name: str,
        import_error_message: Optional[str] = None,
        set_loggers: Optional[Callable] = None,
    ):
        self._module_name = module_name
        self._module = None
        self._import_error_message = import_error_message
        self._set_loggers = set_loggers
        self._lock = threading.RLock()

    def load_module(self):
        # Avoid extra imports when multiple threads try to import the same
        # module. The overhead is minor since import can only run in serial
        # due to GIL even in multi-threaded environments.
        with self._lock:
            if self._module is None:
                try:
                    self._module = importlib.import_module(self._module_name)
                    if self._set_loggers is not None:
                        self._set_loggers()
                except ImportError as e:
                    if self._import_error_message is not None:
                        raise ImportError(self._import_error_message) from e
                    raise
        return self._module

    def __getattr__(self, name: str) -> Any:
        # Attempt to access the attribute, if it fails, assume it's a submodule
        # and lazily import it
        try:
            if name in self.__dict__:
                return self.__dict__[name]
            return getattr(self.load_module(), name)
        except AttributeError:
            # Dynamically create a new LazyImport instance for the submodule
            submodule_name = f'{self._module_name}.{name}'
            lazy_submodule = LazyImport(submodule_name, self._import_error_message)
            setattr(self, name, lazy_submodule)
            return lazy_submodule


def load_lazy_modules(modules: Tuple[LazyImport, ...]):
    """Load lazy modules before entering a function to error out quickly."""

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            for m in modules:
                m.load_module()
            return func(*args, **kwargs)

        return wrapper

    return decorator



import os
import re
from typing import Any, Dict, List, Set

import requests

from konduktor import logging as konduktor_logging
from konduktor.controller import constants

# comma separated list of namespaces to watch for pod errors
WATCHED_NAMESPACES: List[str] = os.environ.get('WATCHED_NAMESPACES', 'default').split(
    ','
)
LOGS_SINCE: int = 10  # retrieves logs generated in the past 10 seconds
LOG_ENDPOINT: str = os.environ.get(
    'LOG_ENDPOINT',
    # this assumes you have access to this endpoint by
    # running as a deployment within the cluster
    # for local development use 'http://localhost:3100' and
    # kubectl port-forward svc/loki -n loki 3100:3100
    'http://loki.loki.svc.cluster.local:3100',
)
QUERY_URL: str = '/loki/api/v1/query_range'

logger = konduktor_logging.get_logger(__name__)


def _query_range(pattern: str, **label_filters) -> List[Dict[str, Any]]:
    """Send LogQL query_range to loki
    https://grafana.com/docs/loki/latest/reference/loki-http-api/#query-logs-within-a-range-of-time

    Args:
        pattern (str): regex pattern to match loglines against

    Returns:
        List[Dict[str, Any]]: List of loglines
    """
    url = f'{LOG_ENDPOINT}{QUERY_URL}'
    formatted_filters = ', '.join(
        f'{key}="{value}"' for key, value in label_filters.items()
    )
    query = r'{' f'{formatted_filters}' r'}' f'|~ {pattern}'
    params = {'query': query, 'since': f'{LOGS_SINCE}s'}
    response = requests.get(url, params=params)
    if response.status_code == 200:
        data = response.json()
        return data['data']['result']
    elif response.status_code == 400:
        logger.error(f'Bad Request: {response.status_code}')
        logger.error(response.json())  # Optionally print the error details
    else:
        logger.error(f'loki query failed {params}')
    return []


def pod_errors() -> Set[str]:
    logger.info('querying pod logs')
    bad_nodes = set()
    for regex in constants.POD_LOG_ERROR_REGEXES:
        for namespace in WATCHED_NAMESPACES:
            log_lines = _query_range(regex, k8s_namespace_name=namespace)
            for line in log_lines:
                log_node = line['stream']['k8s_node_name']
                bad_nodes.add(log_node)
    return bad_nodes


def sxid_error(pattern: str, log_content: str) -> int:
    """Regex pattern match for an xid error, from `log_content` otherwise return 0
    Example Xid error from dmesg
    [1235733.431527] NVRM: Xid (PCI:0000:4e:00): 79, pid='<unknown>', name=<unknown>, GPU has fallen off the bus.
    Example sxid error from dmesg
    [1235733.431527] nvidia-nvswitch3: SXid (PCI:0000:4e:00.0): 12028, Non-fatal, Link 32 egress non-posted PRIV error (First)
    """  # noqa: E501

    match = re.search(pattern, log_content)
    if match:
        return int(match.group(1))

    return 0


def is_sxid_error(log_content: str) -> bool:
    """Returns (S)Xid error code, zero otherwise"""
    error_code = sxid_error(r'SXid.*?: (\d+),', log_content) or sxid_error(
        r'NVRM: Xid.*?: (\d+),', log_content
    )
    return error_code not in constants.ALLOWLISTED_NVSWITCH_SXID_ERRORS


def dmesg_errors() -> Set[str]:
    logger.info('checking dmesg logs')
    pattern = ' or '.join(constants.DMESG_ERROR_REGEXES)
    log_lines = _query_range(pattern, k8s_daemonset_name='dmesg')
    bad_nodes = set()
    for line in log_lines:
        log_node, log_content = line['stream']['k8s_node_name'], line['values'][0][1]
        if is_sxid_error(log_content):
            logger.info(f'node `{log_node}` has (S)Xid error: {log_content}')
        else:
            logger.info(f'dmesg error on node `{log_node}`: {log_content}')
        bad_nodes.add(log_node)
    return bad_nodes


if __name__ == '__main__':
    import time

    while True:
        time.sleep(5)
        print(dmesg_errors())



"""The Konduktor package."""

import os
import subprocess

from konduktor.execution import launch
from konduktor.resource import Resources
from konduktor.task import Task

__all__ = [
    'launch',
    'Resources',
    'Task',
]

# Replaced with the current commit when building the wheels.
_KONDUKTOR_COMMIT_SHA = 'c7becbf4b64d992f56aad47b69770f3345a37297'
os.makedirs(os.path.expanduser('~/.konduktor'), exist_ok=True)


def _get_git_commit():
    if 'KONDUKTOR_COMMIT_SHA' not in _KONDUKTOR_COMMIT_SHA:
        # This is a release build, so we don't need to get the commit hash from
        # git, as it's already been set.
        return _KONDUKTOR_COMMIT_SHA

    # This is a development build (pip install -e .), so we need to get the
    # commit hash from git.
    try:
        cwd = os.path.dirname(__file__)
        commit_hash = subprocess.check_output(
            ['git', 'rev-parse', 'HEAD'],
            cwd=cwd,
            universal_newlines=True,
            stderr=subprocess.DEVNULL,
        ).strip()
        changes = subprocess.check_output(
            ['git', 'status', '--porcelain'],
            cwd=cwd,
            universal_newlines=True,
            stderr=subprocess.DEVNULL,
        ).strip()
        if changes:
            commit_hash += '-dirty'
        return commit_hash
    except Exception:  # pylint: disable=broad-except
        return _KONDUKTOR_COMMIT_SHA


__commit__ = _get_git_commit()
__version__ = '1.0.0.dev0.1.0.dev20250412104322'
__root_dir__ = os.path.dirname(os.path.abspath(__file__))



"""Batch job execution via k8s jobsets
https://jobset.sigs.k8s.io/
https://kueue.sigs.k8s.io/docs/tasks/run/jobsets/
"""

import threading
import time
import typing
from typing import Dict, Optional, Tuple

import colorama

if typing.TYPE_CHECKING:
    import konduktor
    from konduktor.data import storage as storage_lib

from konduktor import config, logging
from konduktor.backends import backend, jobset_utils
from konduktor.utils import kubernetes_utils, loki_utils, rich_utils, ux_utils

Path = str
logger = logging.get_logger(__file__)

POLL_INTERVAL = 5
DEFAULT_ATTACH_TIMEOUT = 300  # 5 minutes
FLUSH_LOGS_TIMEOUT = 5


class JobsetError(Exception):
    pass


def _raise_job_error(job):
    """Checks a jobs conditions and statuses for error"""
    for condition in job.status.conditions:
        if 'ConfigIssue' in condition.message:
            raise ValueError(
                'Job failed with '
                f'{colorama.Style.BRIGHT}{colorama.Fore.RED}'
                f'ConfigIssue: ErrImagePull.{colorama.Style.RESET_ALL} '
                f'Check that your '
                f'{colorama.Style.BRIGHT}{colorama.Fore.YELLOW}'
                f'`image_id`{colorama.Style.RESET_ALL} is correct and '
                f'your container credentials are correct. Image specified '
                f'in your task definition is '
                f'{colorama.Style.BRIGHT}{colorama.Fore.RED}'
                f'`{job.spec.template.spec.containers[0].image}`'
                f'{colorama.Style.RESET_ALL}'
            )
        elif 'BackoffLimitExceeded' == condition.reason:
            raise JobsetError('Job failed with non-zero exit code.')
    logger.error(
        'Job failed with unknown error. Check jobset status in k8s with '
        f'{colorama.Style.BRIGHT}{colorama.Fore.YELLOW}'
        f'`kubectl get job -o yaml {job.metadata.name}`'
        f'{colorama.Style.RESET_ALL}'
    )


def _wait_for_jobset_start(namespace: str, job_name: str):
    time.sleep(2)
    start = time.time()
    timeout = config.get_nested(
        ('kubernetes', 'provision_timeout'),
        default_value=DEFAULT_ATTACH_TIMEOUT,
    )

    while True:
        jobsets = jobset_utils.get_jobset(namespace, job_name)
        assert jobsets is not None, (
            f'Jobset {job_name} ' f'not found in namespace {namespace}'
        )
        if jobsets['status']['replicatedJobsStatus'][0]['ready']:
            logger.info(
                f'task '
                f'{colorama.Fore.CYAN}{colorama.Style.BRIGHT}{job_name}'
                f'{colorama.Style.RESET_ALL} ready'
            )
            break
        elif jobsets['status']['replicatedJobsStatus'][0]['succeeded']:
            return
        elif jobsets['status']['replicatedJobsStatus'][0]['failed']:
            logger.info(
                f'job '
                f'{colorama.Fore.CYAN}{colorama.Style.BRIGHT}{job_name}'
                f'{colorama.Style.RESET_ALL} '
                f'{colorama.Fore.RED}{colorama.Style.BRIGHT}failed{colorama.Style.RESET_ALL}'
            )
            job = jobset_utils.get_job(namespace, job_name)
            _raise_job_error(job)
            return
        if timeout != -1 and time.time() - start > timeout:
            logger.error(
                f'{colorama.Style.BRIGHT}'
                f'{colorama.Fore.RED}Job timed out to schedule.'
                f'{colorama.Style.RESET_ALL}. Deleting job'
            )
            jobset_utils.delete_jobset(namespace, job_name)
            raise JobsetError(
                'Job failed to start within '
                f'timeout of {timeout} seconds. '
                f'Increase or disable timeout '
                f'{colorama.Style.BRIGHT}'
                '`konduktor.provision_timeout: -1`'
                f'{colorama.Style.RESET_ALL}'
            )
        time.sleep(POLL_INTERVAL)


def _wait_for_jobset_completion(namespace: str, job_name: str) -> Tuple[bool, str]:
    while True:
        jobsets = jobset_utils.get_jobset(namespace, job_name)
        assert jobsets is not None, (
            f'Jobset {job_name} ' f'not found in namespace {namespace}'
        )
        if jobsets['status']['replicatedJobsStatus'][0]['succeeded']:
            msg = (
                f'task '
                f'{colorama.Fore.CYAN}{colorama.Style.BRIGHT}{job_name}'
                f'{colorama.Style.RESET_ALL} {colorama.Fore.GREEN}'
                f'{colorama.Style.BRIGHT}finished{colorama.Style.RESET_ALL}'
            )
            return True, msg
        elif jobsets['status']['replicatedJobsStatus'][0]['failed']:
            msg = (
                f'task '
                f'{colorama.Fore.CYAN}{colorama.Style.BRIGHT}{job_name}'
                f'{colorama.Style.RESET_ALL} {colorama.Fore.RED}'
                f'{colorama.Style.BRIGHT}failed{colorama.Style.RESET_ALL}'
            )
            return False, msg
        time.sleep(POLL_INTERVAL)


class JobsetBackend(backend.Backend):
    def _sync_file_mounts(
        self,
        all_file_mounts: Optional[Dict[Path, Path]],
        storage_mounts: Optional[Dict[Path, 'storage_lib.Storage']],
    ) -> None:
        """Syncs files/directories to cloud storage before job launch.

        This uploads any local files/dirs to cloud storage so they can be downloaded
        by the pods when they start.
        """
        pass

    def _sync_workdir(self, workdir: str) -> None:
        """Syncs the working directory to cloud storage before job launch."""

        pass

    def _post_execute(self) -> None:
        """
        TODO(asaiacai): add some helpful messages/commands that a user can run
        to inspect the status of their jobset.
        """
        pass

    def _execute(
        self, task: 'konduktor.Task', detach_run: bool = False, dryrun: bool = False
    ) -> Optional[str]:
        """Executes the task on the cluster. By creating a jobset

        Returns:
            Job id if the task is submitted to the cluster, None otherwise.
        """

        # we should consider just building an image with the cloud provider
        # sdks baked in. These can initialize and pull files first before
        # the working container starts.

        # first define the pod spec then create the jobset definition
        pod_spec = jobset_utils.create_pod_spec(task)
        context = kubernetes_utils.get_current_kube_config_context_name()
        namespace = kubernetes_utils.get_kube_config_context_namespace(context)
        # TODO(asaiacai): need to set env variables in pod
        jobset_utils.create_jobset(
            namespace,
            task,
            pod_spec['kubernetes']['pod_config'],
            dryrun=dryrun,
        )

        if not dryrun and not detach_run:
            with ux_utils.print_exception_no_traceback():
                with rich_utils.safe_status(
                    ux_utils.spinner_message('waiting for job to start\n')
                ):
                    _wait_for_jobset_start(namespace, task.name)
                try:
                    log_thread = threading.Thread(
                        target=loki_utils.tail_loki_logs_ws,
                        args=(task.name,),
                        daemon=True,
                    )
                    logger.info('streaming logs...')
                    log_thread.start()
                    is_success, msg = _wait_for_jobset_completion(namespace, task.name)
                    log_thread.join(timeout=2.0)  # give the job sometime to flush logs
                    if not is_success:
                        logger.error(msg)
                    else:
                        logger.info(msg)
                except KeyboardInterrupt:
                    logger.info('detaching from log stream...')
                except Exception as err:
                    logger.error(
                        f'Check if job resources are '
                        f'active/queued with '
                        f'{colorama.Style.BRIGHT}'
                        f'`konduktor status`'
                        f'{colorama.Style.RESET_ALL}'
                    )
                    raise JobsetError(f'error: {err}')
        else:
            logger.info('detaching from run.')
        return task.name



from konduktor.data.aws import S3CloudStorage
from konduktor.data.gcp import GcsCloudStorage

# Maps bucket's URIs prefix(scheme) to its corresponding storage class

_REGISTRY = {
    'gs': GcsCloudStorage(),
    's3': S3CloudStorage(),
    # TODO(asaiacai): Add other cloud stores here
    # 'r2': R2CloudStorage(),
    # 'cos': IBMCosCloudStorage(),
    # 'oci': OciCloudStorage(),
    # # TODO: This is a hack, as Azure URL starts with https://, we should
    # # refactor the registry to be able to take regex, so that Azure blob can
    # # be identified with `https://(.*?)\.blob\.core\.windows\.net`
    # 'https': AzureBlobCloudStorage()
}

_STORE_ENABLED_CLOUDS = list(_REGISTRY.keys())



import asyncio
import datetime
import os
import time
from typing import Dict, List

import requests
from socketio import AsyncServer  # Import the AsyncServer for ASGI compatibility

from konduktor import logging as konduktor_logging

# SocketIO configuration
socketio = AsyncServer(
    cors_allowed_origins='*', ping_interval=25, ping_timeout=60, async_mode='asgi'
)

logger = konduktor_logging.get_logger(__name__)

# Global variables
CLIENT_CONNECTED = False
FIRST_RUN = True
BACKGROUND_TASK_RUNNING = False
LOG_CHECKPOINT_TIME = None
SELECTED_NAMESPACES: list[str] = []

# "http://loki.loki.svc.cluster.local:3100/loki/api/v1/query_range" for prod
# "http://localhost:3100/loki/api/v1/query_range" for local
LOGS_URL = os.environ.get('LOGS_URL', 'http://localhost:3100/loki/api/v1/query_range')


def format_log_entry(entry: List[str], namespace: str) -> Dict[str, str]:
    """
    Formats a log entry and its corresponding namespace

    Args:
        entry (List[str]): A list of log entry strings to be formatted.
        namespace (str): The namespace to apply to each log entry.

    Returns:
        Dict[str, str]: an object with the following properties:
        timestamp, log (message), and namespace
    """
    timestamp_ns = entry[0]
    log_message = entry[1]
    timestamp_s = int(timestamp_ns) / 1e9
    dt = datetime.datetime.utcfromtimestamp(timestamp_s)
    human_readable_time = dt.strftime('%Y-%m-%d %H:%M:%S')
    formatted_log = {
        'timestamp': human_readable_time,
        'log': log_message,
        'namespace': namespace,
    }
    return formatted_log


def get_logs(FIRST_RUN: bool) -> List[Dict[str, str]]:
    global LOG_CHECKPOINT_TIME

    logger.debug(f'Selected namespaces: {SELECTED_NAMESPACES}')

    # Use the selected namespaces in the query
    namespace_filter = (
        '|'.join(SELECTED_NAMESPACES) if SELECTED_NAMESPACES else 'default'
    )
    query = f'{{k8s_namespace_name=~"{namespace_filter}"}}'

    logger.debug(f'Loki logs query: {query}')

    if FIRST_RUN:
        # Calculate how many nanoseconds to look back when first time looking at logs
        # (currently 1 hour)
        now = int(time.time() * 1e9)
        one_hour_ago = now - int(3600 * 1e9)
        start_time = str(one_hour_ago)
    else:
        # calculate new start_time based on newest, last message
        if LOG_CHECKPOINT_TIME is None:
            LOG_CHECKPOINT_TIME = 0
        start_time = str(int(LOG_CHECKPOINT_TIME) + 1)

    params = {'query': query, 'start': start_time, 'limit': '300'}

    url = LOGS_URL
    response = requests.get(url, params=params)
    formatted_logs = []

    last = 0

    if response.status_code == 200:
        data = response.json()
        rows = data['data']['result']

        for row in rows:
            namespace = row['stream']['k8s_namespace_name']
            for value in row['values']:
                last = max(int(value[0]), last)
                formatted_logs.append(format_log_entry(value, namespace))

    if formatted_logs:
        # sort because sometimes loki API is wrong and logs are out of order
        formatted_logs.sort(
            key=lambda log: datetime.datetime.strptime(
                log['timestamp'], '%Y-%m-%d %H:%M:%S'
            )
        )
        LOG_CHECKPOINT_TIME = last

    logger.debug(f'Formatted logs length: {len(formatted_logs)}')

    return formatted_logs


async def send_logs():
    global CLIENT_CONNECTED, FIRST_RUN, BACKGROUND_TASK_RUNNING
    while CLIENT_CONNECTED:
        logs = get_logs(FIRST_RUN)

        FIRST_RUN = False  # After the first successful fetch, set to False
        if logs:
            await socketio.emit('log_data', logs)

        await asyncio.sleep(5)

    # Background task is no longer running after the loop
    BACKGROUND_TASK_RUNNING = False


@socketio.event
async def connect(sid, environ):
    global CLIENT_CONNECTED, FIRST_RUN, BACKGROUND_TASK_RUNNING
    CLIENT_CONNECTED = True
    FIRST_RUN = True
    logger.debug('Client connected')

    # Start the background task only if it's not already running
    if not BACKGROUND_TASK_RUNNING:
        BACKGROUND_TASK_RUNNING = True
        socketio.start_background_task(send_logs)


@socketio.event
async def update_namespaces(sid, namespaces):
    global SELECTED_NAMESPACES
    SELECTED_NAMESPACES = namespaces
    logger.debug('Updated namespaces')


@socketio.event
async def disconnect(sid):
    global CLIENT_CONNECTED, FIRST_RUN, BACKGROUND_TASK_RUNNING
    CLIENT_CONNECTED = False
    FIRST_RUN = True
    BACKGROUND_TASK_RUNNING = False
    logger.debug('Client disconnected')



# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: skypilot: https://github.com/skypilot-org/skypilot
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Resources: compute requirements of Tasks."""

import functools
from typing import Any, Dict, List, Optional, Union

from konduktor import logging
from konduktor.utils import accelerator_registry, common_utils, schemas, ux_utils

logger = logging.get_logger(__name__)

_DEFAULT_DISK_SIZE_GB = 256


class Resources:
    """Resources: compute requirements of Tasks.

    This class is immutable once created (to ensure some validations are done
    whenever properties change). To update the property of an instance of
    Resources, use `resources.copy(**new_properties)`.

    Used:

    * for representing resource requests for task

    """

    # If any fields changed, increment the version. For backward compatibility,
    # modify the __setstate__ method to handle the old version.
    _VERSION = 1

    def __init__(
        self,
        cloud: Optional[Any] = None,
        cpus: Union[None, int, float, str] = None,
        memory: Union[None, int, float, str] = None,
        accelerators: Optional[str] = None,
        image_id: Union[str, None] = None,
        disk_size: Optional[int] = None,
        labels: Optional[Dict[str, str]] = None,
        # Internal use only.
        # pylint: disable=invalid-name
        _cluster_config_overrides: Optional[Dict[str, Any]] = None,
    ):
        """Initialize a Resources object.

        All fields are optional.  ``Resources.is_launchable`` decides whether
        the Resources is fully specified to launch an instance.

        Examples:
          .. code-block:: python

            # Specifying required resources; the system decides the
            # cloud/instance type. The below are equivalent:
            konduktor.Resources(accelerators='V100')
            konduktor.Resources(accelerators='V100:1')
            konduktor.Resources(accelerators={'V100': 1})
            konduktor.Resources(cpus='2+', memory='16+', accelerators='V100')

        Args:
          cloud: the cloud to use. (deprecated) all jobs are submitted to k8s
          instance_type: the instance type to use.
          cpus: the number of CPUs required for the task.
            If a str, must be a string of the form ``'2'`` or ``'2+'``, where
            the ``+`` indicates that the task requires at least 2 CPUs.
          memory: the amount of memory in GiB required. If a
            str, must be a string of the form ``'16'`` or ``'16+'``, where
            the ``+`` indicates that the task requires at least 16 GB of memory.
          accelerators: the accelerators required. If a str, must be
            a string of the form ``'V100'`` or ``'V100:2'``, where the ``:2``
            indicates that the task requires 2 V100 GPUs. If a dict, must be a
            dict of the form ``{'V100': 2}`` or ``{'tpu-v2-8': 1}``.

          image_id: docker image to use

          disk_size: the size of the OS disk in GiB.
          labels: the labels to apply to the instance. These are useful for
            assigning metadata that may be used by external tools.
            Implementation depends on the chosen cloud - On AWS, labels map to
            instance tags. On GCP, labels map to instance labels. On
            Kubernetes, labels map to pod labels. On other clouds, labels are
            not supported and will be ignored.
        Raises:
            ValueError: if some attributes are invalid.
            exceptions.NoCloudAccessError: if no public cloud is enabled.
        """
        self._version = self._VERSION
        if cloud is not None:
            raise ValueError('cloud specified, but all jobs are submitted to k8s')
        self._cloud = cloud

        if disk_size is not None:
            if round(disk_size) != disk_size:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'OS disk size must be an integer. Got: {disk_size}.'
                    )
            self._disk_size = int(disk_size)
        else:
            self._disk_size = _DEFAULT_DISK_SIZE_GB

        # self._image_id is a dict of {region: image_id}.
        # The key is None if the same image_id applies for all regions.
        self._image_id = image_id
        if isinstance(image_id, str):
            self._image_id = image_id.strip()

        self._labels = labels
        self._cluster_config_overrides = _cluster_config_overrides

        self._set_cpus(cpus)
        self._set_memory(memory)
        self._set_accelerators(accelerators)

        # TODO: move these out of init to prevent repeated calls.
        self._try_validate_cpus_mem()
        self._try_validate_image_id()

    def __repr__(self) -> str:
        """Returns a string representation for display.

        Examples:

            >>> konduktor.Resources(accelerators='V100')
            <Kubernetes>({'V100': 1})

        """
        accelerators = ''
        if self.accelerators is not None:
            accelerators = f', {self.accelerators}'

        cpus = ''
        if self._cpus is not None:
            cpus = f', cpus={self._cpus}'

        memory = ''
        if self.memory is not None:
            memory = f', mem={self.memory}'

        image_id = ''
        if self.image_id is not None:
            image_id = f', image_id={self.image_id}'
        else:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'no image id for the task was specified. You must '
                    'specify an image id for this task (e.g. '
                    '`nvcr.io/nvidia/pytorch:xx.xx-py3`'
                )

        disk_size = ''
        if self.disk_size != _DEFAULT_DISK_SIZE_GB:
            disk_size = f', disk_size={self.disk_size}'

        # Do not show region/zone here as `konduktor status -a` would show them as
        # separate columns. Also, Resources repr will be printed during
        # failover, and the region may be dynamically determined.
        hardware_str = f'{cpus}{memory}{accelerators}{image_id}' f'{disk_size}'
        # It may have leading ',' (for example, instance_type not set) or empty
        # spaces.  Remove them.
        while hardware_str and hardware_str[0] in (',', ' '):
            hardware_str = hardware_str[1:]

        return f'({hardware_str})'

    @property
    def cloud(self):
        return self._cloud

    @property
    @functools.lru_cache(maxsize=1)
    def cpus(self) -> Optional[str]:
        """Returns the number of vCPUs that each instance must have.

        For example, cpus='4' means each instance must have exactly 4 vCPUs,
        and cpus='4+' means each instance must have at least 4 vCPUs.

        (Developer note: The cpus field is only used to select the instance type
        at launch time. Thus, Resources in the backend's ResourceHandle will
        always have the cpus field set to None.)
        """
        if self._cpus is not None:
            return self._cpus
        return None

    @property
    def memory(self) -> Optional[str]:
        """Returns the memory that each instance must have in GB.

        For example, memory='16' means each instance must have exactly 16GB
        memory; memory='16+' means each instance must have at least 16GB
        memory.

        (Developer note: The memory field is only used to select the instance
        type at launch time. Thus, Resources in the backend's ResourceHandle
        will always have the memory field set to None.)
        """
        return self._memory

    @property
    @functools.lru_cache(maxsize=1)
    def accelerators(self) -> Optional[Dict[str, int]]:
        """Returns the accelerators field directly or by inferring.

        For example, Resources(AWS, 'p3.2xlarge') has its accelerators field
        set to None, but this function will infer {'V100': 1} from the instance
        type.
        """
        if self._accelerators is not None:
            return self._accelerators
        return None

    @property
    def disk_size(self) -> int:
        return self._disk_size

    @property
    def image_id(self) -> Optional[str]:
        return self._image_id

    @property
    def labels(self) -> Optional[Dict[str, str]]:
        return self._labels

    @property
    def cluster_config_overrides(self) -> Dict[str, Any]:
        if self._cluster_config_overrides is None:
            return {}
        return self._cluster_config_overrides

    def _set_cpus(
        self,
        cpus: Union[None, int, float, str],
    ) -> None:
        if cpus is None:
            self._cpus = None
            return

        self._cpus = str(cpus)
        if isinstance(cpus, str):
            if cpus.endswith('+'):
                num_cpus_str = cpus[:-1]
            else:
                num_cpus_str = cpus

            try:
                num_cpus = float(num_cpus_str)
            except ValueError:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'The "cpus" field should be either a number or '
                        f'a string "<number>+". Found: {cpus!r}'
                    ) from None
        else:
            num_cpus = float(cpus)

        if num_cpus <= 0:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'The "cpus" field should be positive. Found: {cpus!r}'
                )

    def _set_memory(
        self,
        memory: Union[None, int, float, str],
    ) -> None:
        if memory is None:
            self._memory = None
            return

        self._memory = str(memory)
        if isinstance(memory, str):
            if memory.endswith(('+', 'x')):
                # 'x' is used internally for make sure our resources used by
                # jobs controller (memory: 3x) to have enough memory based on
                # the vCPUs.
                num_memory_gb = memory[:-1]
            else:
                num_memory_gb = memory

            try:
                memory_gb = float(num_memory_gb)
            except ValueError:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'The "memory" field should be either a number or '
                        f'a string "<number>+". Found: {memory!r}'
                    ) from None
        else:
            memory_gb = float(memory)

        if memory_gb <= 0:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'The "cpus" field should be positive. Found: {memory!r}'
                )

    def _set_accelerators(
        self,
        accelerators: Union[None, str, Dict[str, int]],
        accelerator_args: Optional[Dict[str, str]] = None,
    ) -> None:
        """Sets accelerators.

        Args:
            accelerators: A string or a dict of accelerator types to counts.
            accelerator_args: (deprecated) A dict of accelerator types to args.
        """
        if accelerators is not None:
            if isinstance(accelerators, str):  # Convert to Dict[str, int].
                if ':' not in accelerators:
                    accelerators = {accelerators: 1}
                else:
                    splits = accelerators.split(':')
                    parse_error = (
                        'The "accelerators" field as a str '
                        'should be <name> or <name>:<cnt>. '
                        f'Found: {accelerators!r}'
                    )
                    if len(splits) != 2:
                        with ux_utils.print_exception_no_traceback():
                            raise ValueError(parse_error)
                    try:
                        num = float(splits[1])
                        num = int(num)
                        accelerators = {splits[0]: num}
                    except ValueError:
                        with ux_utils.print_exception_no_traceback():
                            raise ValueError(parse_error) from None

            # Canonicalize the accelerator names.
            accelerators = {
                accelerator_registry.canonicalize_accelerator_name(acc): acc_count
                for acc, acc_count in accelerators.items()
            }

            acc, _ = list(accelerators.items())[0]

        self._accelerators = accelerators

    def _try_validate_cpus_mem(self) -> None:
        """Try to validate the cpus and memory attributes.

        Raises:
            ValueError: if the attributes are invalid.
        """
        if self._cpus is None and self._memory is None:
            return

    def _try_validate_image_id(self) -> None:
        """Try to validate the image_id attribute.

        Raises:
            ValueError: if the attribute is invalid.
        """
        if self._image_id is None:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'no image id for the task was specified. You must '
                    'specify an image id for this task (e.g. '
                    '`nvcr.io/nvidia/pytorch:xx.xx-py3`'
                )

    def get_accelerators_str(self) -> str:
        accelerators = self.accelerators
        accel_str = ''
        if accelerators is None:
            accel_str = '-'
        elif isinstance(accelerators, dict) and len(accelerators) == 1:
            accel_name, accel_count = list(accelerators.items())[0]
            accel_str = f'{accel_name}:{accel_count}'
        return accel_str

    def copy(self, **override) -> 'Resources':
        """Returns a copy of the given Resources."""
        resources = Resources(
            cloud=override.pop('cloud', self.cloud),
            cpus=override.pop('cpus', self._cpus),
            memory=override.pop('memory', self.memory),
            accelerators=override.pop('accelerators', self.accelerators),
            disk_size=override.pop('disk_size', self.disk_size),
            image_id=override.pop('image_id', self.image_id),
            labels=override.pop('labels', self.labels),
        )
        assert len(override) == 0
        return resources

    @classmethod
    def from_yaml_config(cls, config: Optional[Dict[str, Any]]) -> 'Resources':
        if config is None:
            return Resources()
        common_utils.validate_schema(
            config, schemas.get_resources_schema(), 'Invalid resources YAML: '
        )

        def _override_resources(
            base_resource_config: Dict[str, Any], override_configs: List[Dict[str, Any]]
        ) -> List[Resources]:
            resources_list = []
            for override_config in override_configs:
                new_resource_config = base_resource_config.copy()
                # Labels are handled separately.
                override_labels = override_config.pop('labels', None)
                new_resource_config.update(override_config)

                # Update the labels with the override labels.
                labels = new_resource_config.get('labels', None)
                if labels is not None and override_labels is not None:
                    labels.update(override_labels)
                elif override_labels is not None:
                    labels = override_labels
                new_resource_config['labels'] = labels

                # Call from_yaml_config again instead of
                # _from_yaml_config_single to handle the case, where both
                # multiple accelerators and `any_of` is specified.
                # This will not cause infinite recursion because we have made
                # sure that `any_of` and `ordered` cannot be specified in the
                # resource candidates in `any_of` or `ordered`, by the schema
                # validation above.
                resources_list.extend([Resources.from_yaml_config(new_resource_config)])

            return resources_list

        config = config.copy()

        return Resources._from_yaml_config_single(config)

    @classmethod
    def _from_yaml_config_single(cls, config: Dict[str, str]) -> 'Resources':
        resources_fields: Dict[str, Any] = {}
        resources_fields['cpus'] = config.pop('cpus', None)
        resources_fields['memory'] = config.pop('memory', None)
        resources_fields['accelerators'] = config.pop('accelerators', None)
        resources_fields['disk_size'] = config.pop('disk_size', None)
        resources_fields['image_id'] = config.pop('image_id', None)
        resources_fields['labels'] = config.pop('labels', None)

        if resources_fields['cpus'] is not None:
            resources_fields['cpus'] = str(resources_fields['cpus'])
        if resources_fields['memory'] is not None:
            resources_fields['memory'] = str(resources_fields['memory'])
        # TODO(asaiacai): should we remove disk size
        # since we aren't letting users set this at the host level?
        if resources_fields['disk_size'] is not None:
            resources_fields['disk_size'] = int(resources_fields['disk_size'])

        assert not config, f'Invalid resource args: {config.keys()}'
        return Resources(**resources_fields)

    def to_yaml_config(self) -> Dict[str, Union[str, int]]:
        """Returns a yaml-style dict of config for this resource bundle."""
        config = {}

        def add_if_not_none(key, value):
            if value is not None and value != 'None':
                config[key] = value

        add_if_not_none('cloud', str(self.cloud))
        add_if_not_none('cpus', self._cpus)
        add_if_not_none('memory', self.memory)
        add_if_not_none('accelerators', self.accelerators)

        add_if_not_none('disk_size', self.disk_size)
        add_if_not_none('image_id', self.image_id)
        add_if_not_none('labels', self.labels)
        return config



# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: skypilot: https://github.com/skypilot-org/skypilot
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Kubernetes utilities."""

import functools
import math
import os
import re
import typing
from typing import Any, Dict, List, Optional, Tuple, Union

import kubernetes
import yaml

from konduktor import config, kube_client, logging
from konduktor.utils import common_utils, kubernetes_enums

if typing.TYPE_CHECKING:
    pass

DEFAULT_NAMESPACE = 'default'

DEFAULT_SERVICE_ACCOUNT_NAME = 'konduktor-service-account'

MEMORY_SIZE_UNITS = {
    'B': 1,
    'K': 2**10,
    'M': 2**20,
    'G': 2**30,
    'T': 2**40,
    'P': 2**50,
}

# The resource keys used by Kubernetes to track NVIDIA GPUs and Google on
# nodes. These keys are typically used in the node's status.allocatable
# or status.capacity fields to indicate the available resources on the node.
GPU_RESOURCE_KEY = 'nvidia.com/gpu'

NO_ACCELERATOR_HELP_MESSAGE = (
    'If your cluster contains GPUs, make sure '
    f'{GPU_RESOURCE_KEY} resource is available '
    'on the nodes and the node labels for identifying GPUs '
    '(e.g. `nvidia.com/gpu` are setup correctly. '
)


logger = logging.get_logger(__name__)


class GPULabelFormatter:
    """Base class to define a GPU label formatter for a Kubernetes cluster

    A GPU label formatter is a class that defines how to use GPU type labels in
    a Kubernetes cluster. It is used by the Kubernetes cloud class to pick the
    key:value pair to use as node selector for GPU nodes.
    """

    @classmethod
    def get_label_key(cls, accelerator: Optional[str] = None) -> str:
        """Returns the label key for GPU type used by the Kubernetes cluster"""
        raise NotImplementedError

    @classmethod
    def get_label_keys(cls) -> List[str]:
        """Returns a list of label keys for GPU used by Kubernetes cluster."""
        raise NotImplementedError

    @classmethod
    def get_label_value(cls, accelerator: str) -> str:
        """Given a GPU type, returns the label value to be used"""
        raise NotImplementedError

    @classmethod
    def match_label_key(cls, label_key: str) -> bool:
        """Checks if the given label key matches the formatter's label keys"""
        raise NotImplementedError

    @classmethod
    def get_accelerator_from_label_value(cls, value: str) -> str:
        """Given a label value, returns the GPU type"""
        raise NotImplementedError

    @classmethod
    def validate_label_value(cls, value: str) -> Tuple[bool, str]:
        """Validates if the specified label value is correct.

        Used to check if the labelling on the cluster is correct and
        preemptively raise an error if it is not.

        Returns:
            bool: True if the label value is valid, False otherwise.
            str: Error message if the label value is invalid, None otherwise.
        """
        del value
        return True, ''


def get_gke_accelerator_name(accelerator: str) -> str:
    """Returns the accelerator name for GKE clusters.

    Uses the format - nvidia-tesla-<accelerator>.
    A100-80GB, H100-80GB, L4 are an exception. They use nvidia-<accelerator>.
    types are an exception as well keeping the given name.
    """
    if accelerator == 'H100':
        # H100 is named as H100-80GB in GKE.
        accelerator = 'H100-80GB'
    if accelerator in ('A100-80GB', 'L4', 'H100-80GB', 'H100-MEGA-80GB'):
        # A100-80GB, L4, H100-80GB and H100-MEGA-80GB
        # have a different name pattern.
        return 'nvidia-{}'.format(accelerator.lower())
    else:
        return 'nvidia-tesla-{}'.format(accelerator.lower())


class GKELabelFormatter(GPULabelFormatter):
    """GKE label formatter

    GKE nodes by default are populated with `cloud.google.com/gke-accelerator`
    label, which is used to identify the GPU type.
    """

    GPU_LABEL_KEY = 'cloud.google.com/gke-accelerator'
    ACCELERATOR_COUNT_LABEL_KEY = 'cloud.google.com/gke-accelerator-count'

    @classmethod
    def get_label_key(cls, accelerator: Optional[str] = None) -> str:
        return cls.GPU_LABEL_KEY

    @classmethod
    def get_label_keys(cls) -> List[str]:
        return [cls.GPU_LABEL_KEY]

    @classmethod
    def match_label_key(cls, label_key: str) -> bool:
        return label_key in cls.get_label_keys()

    @classmethod
    def get_label_value(cls, accelerator: str) -> str:
        return get_gke_accelerator_name(accelerator)

    @classmethod
    def get_accelerator_from_label_value(cls, value: str) -> str:
        if value.startswith('nvidia-tesla-'):
            return value.replace('nvidia-tesla-', '').upper()
        elif value.startswith('nvidia-'):
            acc = value.replace('nvidia-', '').upper()
            if acc == 'H100-80GB':
                # H100 can be either H100-80GB or H100-MEGA-80GB in GKE
                # we map H100 ---> H100-80GB and keep H100-MEGA-80GB
                # to distinguish between a3-high and a3-mega instances
                return 'H100'
            return acc
        else:
            raise ValueError(f'Invalid accelerator name in GKE cluster: {value}')


class GFDLabelFormatter(GPULabelFormatter):
    """GPU Feature Discovery label formatter

    NVIDIA GPUs nodes are labeled by GPU feature discovery
    e.g. nvidia.com/gpu.product=NVIDIA-H100-80GB-HBM3
    https://github.com/NVIDIA/gpu-feature-discovery

    GPU feature discovery is included as part of the
    NVIDIA GPU Operator:
    https://docs.nvidia.com/datacenter/cloud-native/gpu-operator/latest/overview.html

    This LabelFormatter can't be used in autoscaling clusters since accelerators
    may map to multiple label, so we're not implementing `get_label_value`
    """

    LABEL_KEY = 'nvidia.com/gpu.product'

    @classmethod
    def get_label_key(cls, accelerator: Optional[str] = None) -> str:
        return cls.LABEL_KEY

    @classmethod
    def get_label_keys(cls) -> List[str]:
        return [cls.LABEL_KEY]

    @classmethod
    def get_label_value(cls, accelerator: str) -> str:
        """An accelerator can map to many Nvidia GFD labels
        (e.g., A100-80GB-PCIE vs. A100-SXM4-80GB).
        As a result, we do not support get_label_value for GFDLabelFormatter."""
        raise NotImplementedError

    @classmethod
    def match_label_key(cls, label_key: str) -> bool:
        return label_key == cls.LABEL_KEY

    @classmethod
    def get_accelerator_from_label_value(cls, value: str) -> str:
        """Searches against a canonical list of NVIDIA GPUs and pattern
        matches the canonical GPU name against the GFD label.
        """
        canonical_gpu_names = [
            'A100-80GB',
            'A100',
            'A10G',
            'H100',
            'K80',
            'M60',
            'T4g',
            'T4',
            'V100',
            'A10',
            'P4000',
            'P100',
            'P40',
            'P4',
            'L40',
            'L4',
        ]
        for canonical_name in canonical_gpu_names:
            # A100-80G accelerator is A100-SXM-80GB or A100-PCIE-80GB
            if canonical_name == 'A100-80GB' and re.search(r'A100.*-80GB', value):
                return canonical_name
            # Use word boundary matching to prevent substring matches
            elif re.search(rf'\b{re.escape(canonical_name)}\b', value):
                return canonical_name

        # If we didn't find a canonical name:
        # 1. remove 'NVIDIA-' (e.g., 'NVIDIA-RTX-A6000' -> 'RTX-A6000')
        # 2. remove 'GEFORCE-' (e.g., 'NVIDIA-GEFORCE-RTX-3070' -> 'RTX-3070')
        # 3. remove 'RTX-' (e.g. 'RTX-6000' -> 'RTX6000')
        return (
            value.upper()
            .replace('NVIDIA-', '')
            .replace('GEFORCE-', '')
            .replace('RTX-', 'RTX')
        )


# LABEL_FORMATTER_REGISTRY stores the label formats that will try to
# discover the accelerator type from. The order of the list is important, as
# it will be used to determine the priority of the label formats when
# auto-detecting the GPU label type.
LABEL_FORMATTER_REGISTRY = [GKELabelFormatter, GFDLabelFormatter]

# Mapping of autoscaler type to label formatter
AUTOSCALER_TO_LABEL_FORMATTER = {
    kubernetes_enums.KubernetesAutoscalerType.GKE: GKELabelFormatter,
}


@functools.lru_cache()
def get_current_kube_config_context_name() -> Optional[str]:
    """Get the current kubernetes context from the kubeconfig file

    Returns:
        str | None: The current kubernetes context if it exists, None otherwise
    """
    k8s = kubernetes
    try:
        _, current_context = k8s.config.list_kube_config_contexts()
        return current_context['name']
    except k8s.config.config_exception.ConfigException:
        return None


@functools.lru_cache(maxsize=10)
def get_kubernetes_nodes(context: Optional[str] = None) -> List[Any]:
    """Gets the kubernetes nodes in the context.

    If context is None, gets the nodes in the current context.
    """
    if context is None:
        context = get_current_kube_config_context_name()

    nodes = (
        kube_client.core_api(context)
        .list_node(_request_timeout=kubernetes.API_TIMEOUT)
        .items
    )
    return nodes


@functools.lru_cache()
def detect_gpu_label_formatter(
    context: Optional[str],
) -> Tuple[Optional[GPULabelFormatter], Dict[str, List[Tuple[str, str]]]]:
    """Detects the GPU label formatter for the Kubernetes cluster

    Returns:
        GPULabelFormatter: The GPU label formatter for the cluster, if found.
        Dict[str, List[Tuple[str, str]]]: A mapping of nodes and the list of
             labels on each node. E.g., {'node1': [('label1', 'value1')]}
    """
    # Get all labels across all nodes
    node_labels: Dict[str, List[Tuple[str, str]]] = {}
    nodes = get_kubernetes_nodes(context)
    for node in nodes:
        node_labels[node.metadata.name] = []
        for label, value in node.metadata.labels.items():
            node_labels[node.metadata.name].append((label, value))

    label_formatter = None

    # Check if the node labels contain any of the GPU label prefixes
    for lf in LABEL_FORMATTER_REGISTRY:
        for _, label_list in node_labels.items():
            for label, _ in label_list:
                if lf.match_label_key(label):
                    label_formatter = lf()
                    return label_formatter, node_labels

    return label_formatter, node_labels


@functools.lru_cache()
def get_kube_config_context_namespace(context_name: Optional[str] = None) -> str:
    """Get the current kubernetes context namespace from the kubeconfig file

    Returns:
        str | None: The current kubernetes context namespace if it exists, else
            the default namespace.
    """
    k8s = kubernetes
    ns_path = '/var/run/secrets/kubernetes.io/serviceaccount/namespace'
    # If using in-cluster context, get the namespace from the service account
    # namespace file. Uses the same logic as adaptors.kubernetes._load_config()
    # to stay consistent with in-cluster config loading.
    if context_name == kube_client.in_cluster_context_name() or context_name is None:
        if os.path.exists(ns_path):
            with open(ns_path, encoding='utf-8') as f:
                return f.read().strip()
    # If not in-cluster, get the namespace from kubeconfig
    try:
        contexts, current_context = k8s.config.list_kube_config_contexts()
        if context_name is None:
            context = current_context
        else:
            context = next((c for c in contexts if c['name'] == context_name), None)
            if context is None:
                return DEFAULT_NAMESPACE

        if 'namespace' in context['context']:
            return context['context']['namespace']
        else:
            return DEFAULT_NAMESPACE
    except k8s.config.config_exception.ConfigException:
        return DEFAULT_NAMESPACE


def check_credentials(
    context: Optional[str], timeout: int = kube_client.API_TIMEOUT
) -> Tuple[bool, Optional[str]]:
    """Check if the credentials in kubeconfig file are valid

    Args:
        context (Optional[str]): The Kubernetes context to use. If none, uses
            in-cluster auth to check credentials, if available.
        timeout (int): Timeout in seconds for the test API call

    Returns:
        bool: True if credentials are valid, False otherwise
        str: Error message if credentials are invalid, None otherwise
    """
    try:
        namespace = get_kube_config_context_namespace(context)
        kube_client.core_api(context).list_namespaced_pod(
            namespace, _request_timeout=timeout
        )
        return True, None
    except ImportError:
        return False, (
            '`kubernetes` package is not installed. '
            'Install it with: pip install kubernetes'
        )
    except kube_client.api_exception() as e:
        # Check if the error is due to invalid credentials
        if e.status == 401:
            return (
                False,
                'Invalid credentials - do you have permission '
                'to access the cluster?',
            )
        else:
            return False, f'Failed to communicate with the cluster: {str(e)}'
    except kube_client.config_exception() as e:
        return False, f'Invalid configuration file: {str(e)}'
    except kube_client.max_retry_error():
        return False, (
            'Failed to communicate with the cluster - timeout. '
            'Check if your cluster is running and your network '
            'is stable.'
        )
    except ValueError as e:
        return False, common_utils.format_exception(e)
    except Exception as e:  # pylint: disable=broad-except
        return False, (
            'An error occurred: '
            f'{common_utils.format_exception(e, use_bracket=True)}'
        )


def parse_cpu_or_gpu_resource(resource_qty_str: str) -> Union[int, float]:
    resource_str = str(resource_qty_str)
    if resource_str[-1] == 'm':
        # For example, '500m' rounds up to 1.
        return math.ceil(int(resource_str[:-1]) / 1000)
    else:
        return float(resource_str)


def parse_memory_resource(resource_qty_str: str, unit: str = 'B') -> Union[int, float]:
    """Returns memory size in chosen units given a resource quantity string."""
    if unit not in MEMORY_SIZE_UNITS:
        valid_units = ', '.join(MEMORY_SIZE_UNITS.keys())
        raise ValueError(f'Invalid unit: {unit}. Valid units are: {valid_units}')

    resource_str = str(resource_qty_str)
    bytes_value: Union[int, float]
    try:
        bytes_value = int(resource_str)
    except ValueError:
        memory_size = re.sub(r'([KMGTPB]+)', r' \1', resource_str)
        number, unit_index = [item.strip() for item in memory_size.split()]
        unit_index = unit_index[0]
        bytes_value = float(number) * MEMORY_SIZE_UNITS[unit_index]
    return bytes_value / MEMORY_SIZE_UNITS[unit]


def combine_pod_config_fields(
    cluster_yaml_path: str,
    cluster_config_overrides: Dict[str, Any],
) -> None:
    """Adds or updates fields in the YAML with fields from the ~/.konduktor/config's
    kubernetes.pod_spec dict.
    This can be used to add fields to the YAML that are not supported by
    yet, or require simple configuration (e.g., adding an
    imagePullSecrets field).
    Note that new fields are added and existing ones are updated. Nested fields
    are not completely replaced, instead their objects are merged. Similarly,
    if a list is encountered in the config, it will be appended to the
    destination list.
    For example, if the YAML has the following:
        ```
        ...
        node_config:
            spec:
                containers:
                    - name: ray
                    image: rayproject/ray:nightly
        ```
    and the config has the following:
        ```
        kubernetes:
            pod_config:
                spec:
                    imagePullSecrets:
                        - name: my-secret
        ```
    then the resulting YAML will be:
        ```
        ...
        node_config:
            spec:
                containers:
                    - name: ray
                    image: rayproject/ray:nightly
                imagePullSecrets:
                    - name: my-secret
        ```
    """
    with open(cluster_yaml_path, 'r', encoding='utf-8') as f:
        yaml_content = f.read()
    yaml_obj = yaml.safe_load(yaml_content)
    # We don't use override_configs in `konduktor_config.get_nested`, as merging
    # the pod config requires special handling.
    kubernetes_config = config.get_nested(
        ('kubernetes', 'pod_config'), default_value={}, override_configs={}
    )
    override_pod_config = cluster_config_overrides.get('kubernetes', {}).get(
        'pod_config', {}
    )
    config.merge_k8s_configs(override_pod_config, kubernetes_config)

    yaml_obj['kubernetes']['pod_config'] = override_pod_config

    # Write the updated YAML back to the file
    common_utils.dump_yaml(cluster_yaml_path, yaml_obj)


def combine_metadata_fields(cluster_yaml_path: str) -> None:
    """Updates the metadata for all Kubernetes objects created with
    fields from the ~/.konduktor/config's kubernetes.custom_metadata dict.

    Obeys the same add or update semantics as combine_pod_config_fields().
    """

    with open(cluster_yaml_path, 'r', encoding='utf-8') as f:
        yaml_content = f.read()
    yaml_obj = yaml.safe_load(yaml_content)
    custom_metadata = config.get_nested(('kubernetes', 'custom_metadata'), {})

    # List of objects in the cluster YAML to be updated
    combination_destinations = [
        # Service accounts
        yaml_obj['provider']['autoscaler_service_account']['metadata'],
        yaml_obj['provider']['autoscaler_role']['metadata'],
        yaml_obj['provider']['autoscaler_role_binding']['metadata'],
        yaml_obj['provider']['autoscaler_service_account']['metadata'],
        # Pod spec
        yaml_obj['available_node_types']['ray_head_default']['node_config']['metadata'],
        # Services for pods
        *[svc['metadata'] for svc in yaml_obj['provider']['services']],
    ]

    for destination in combination_destinations:
        config.merge_k8s_configs(custom_metadata, destination)

    # Write the updated YAML back to the file
    common_utils.dump_yaml(cluster_yaml_path, yaml_obj)


def merge_custom_metadata(original_metadata: Dict[str, Any]) -> None:
    """Merges original metadata with custom_metadata from config

    Merge is done in-place, so return is not required
    """
    custom_metadata = config.get_nested(('kubernetes', 'custom_metadata'), {})
    config.merge_k8s_configs(custom_metadata, original_metadata)


def check_nvidia_runtime_class(context: Optional[str] = None) -> bool:
    """Checks if the 'nvidia' RuntimeClass exists in the cluster"""
    # Fetch the list of available RuntimeClasses
    runtime_classes = kube_client.node_api(context).list_runtime_class()

    # Check if 'nvidia' RuntimeClass exists
    nvidia_exists = any(rc.metadata.name == 'nvidia' for rc in runtime_classes.items)
    return nvidia_exists


def check_secret_exists(
    secret_name: str, namespace: str, context: Optional[str]
) -> Tuple[bool, Union[str, Dict[str, Any]]]:
    """Checks if a secret exists in a namespace

    Args:
        secret_name: Name of secret to check
        namespace: Namespace to check
    Returns:
        bool: True if the secret exists, False otherwise
        str: response payload if True, error string otherwise
    """

    try:
        response = kube_client.core_api(context).read_namespaced_secret(
            secret_name, namespace, _request_timeout=kube_client.API_TIMEOUT
        )
    except kube_client.api_exception() as e:
        if e.status == 404:
            return False, str(e)
        raise
    else:
        return True, response


def set_secret(
    secret_name: str,
    namespace: str,
    context: Optional[str],
    secret_key: str,
    secret_value: str,
) -> Tuple[bool, Optional[str]]:
    """
    Create/update a secret in a namespace. Values are encoded to base64.
    """
    secret_exists, response = check_secret_exists(
        secret_name=secret_name,
        namespace=namespace,
        context=context,
    )

    secret_metadata = {'name': secret_name, 'labels': {'parent': 'konduktor'}}
    custom_metadata = config.get_nested(('kubernetes', 'custom_metadata'), {})
    config.merge_k8s_configs(secret_metadata, custom_metadata)

    secret = kubernetes.client.V1Secret(
        metadata=kubernetes.client.V1ObjectMeta(**secret_metadata),
        type='Opaque',
        data={secret_key: secret_value},
    )

    try:
        if secret_exists:
            kube_client.core_api(context).patch_namespaced_secret(
                secret_name, namespace, secret
            )
        else:
            kube_client.core_api(context).create_namespaced_secret(namespace, secret)
    except kube_client.api_exception() as e:
        return False, str(e)
    else:
        logger.debug(
            f'Secret {secret_name} in namespace {namespace} '
            f'in context {context} created/updated'
        )
        return True, None


def get_autoscaler_type() -> Optional[kubernetes_enums.KubernetesAutoscalerType]:
    """Returns the autoscaler type by reading from config"""
    autoscaler_type = config.get_nested(('kubernetes', 'autoscaler'), None)
    if autoscaler_type is not None:
        autoscaler_type = kubernetes_enums.KubernetesAutoscalerType(autoscaler_type)
    return autoscaler_type


# TODO(asaiacai): some checks here for CRDs for jobset and Kueue CRDs, queues, etc.
def is_label_valid(label_key: str, label_value: str) -> Tuple[bool, Optional[str]]:
    # Kubernetes labels can be of the format <domain>/<key>: <value>
    key_regex = re.compile(
        # Look-ahead to ensure proper domain formatting up to a slash
        r'^(?:(?=[a-z0-9]([-a-z0-9.]*[a-z0-9])?\/)'
        # Match domain: starts and ends with alphanum up to 253 chars
        # including a slash in the domain.
        r'[a-z0-9]([-a-z0-9.]{0,251}[a-z0-9])?\/)?'
        # Match key: starts and ends with alphanum, upto to 63 chars.
        r'[a-z0-9]([-a-z0-9_.]{0,61}[a-z0-9])?$'
    )
    value_regex = re.compile(r'^([a-zA-Z0-9]([-a-zA-Z0-9_.]{0,61}[a-zA-Z0-9])?)?$')
    key_valid = bool(key_regex.match(label_key))
    value_valid = bool(value_regex.match(label_value))
    error_msg = None
    condition_msg = (
        'Value must consist of alphanumeric characters or '
        "'-', '_', '.', and must be no more than 63 "
        'characters in length.'
    )
    if not key_valid:
        error_msg = f'Invalid label key {label_key} for Kubernetes. ' f'{condition_msg}'
    if not value_valid:
        error_msg = (
            f'Invalid label value {label_value} for Kubernetes. ' f'{condition_msg}'
        )
    if not key_valid or not value_valid:
        return False, error_msg
    return True, None



from typing import List

import kubernetes

from konduktor import kube_client
from konduktor import logging as konduktor_logging

# node taint/label
NODE_HEALTH_LABEL = 'trainy.konduktor.ai/faulty'

logger = konduktor_logging.get_logger(__name__)


def nccl_single_test(node: str, thresh: int = 400):
    """Runs NCCL test within a node. Tests NVLINK BW

    Args:
        node (str): k8s node name
        thresh (int, optional): minimum busbw to be considered healthy
        H100SXM should report 450GB/s max theoretically. Default 400

    """


def nccl_pair_test(nodeA: str, nodeB: str, thresh: int = 350):
    """Runs NCCL test between a pair of nodes. Tests
    internode bandwidth

    Args:
        nodeA (str): k8s node name
        nodeB (str): k8s node name
        thresh (int, optional): minimum busbw to be considered healthy
        ConnectX-7 cards have a theoretical max BW of 400GB/s
        Defaults to 350GB/s.
    """
    pass


def health_check():
    """Gathers nodes with label/taint `trainy.konduktor.ai/faulty=true:NoSchedule`
    and attempts to run NCCL test on them. Nodes that pass
    have their label/taint removed.
    """
    pass


def untaint(node_name: str):
    """Removes label/taint of `trainy.konduktor.ai/faulty=true:NoSchedule`

    Args:
        node (str): k8s node name
    """
    core_api = kube_client.core_api()
    node = core_api.read_node(
        name=node_name,
        _request_timeout=kube_client.API_TIMEOUT,
    )

    if node.spec.taints is not None:
        node.spec.taints = [
            taint for taint in node.spec.taints if taint.key != NODE_HEALTH_LABEL
        ]

    # Patch the node with the new taints
    core_api.patch_node(
        name=node_name,
        body=node,
        _request_timeout=kube_client.API_TIMEOUT,
    )

    logger.info(f'Node {node_name} taint removed.')


def taint(node_name: str):
    """Labels/Taints node with `trainy.konduktor.ai/faulty=true:NoSchedule`

    Args:
        node (str): k8s node name
    """
    core_api = kube_client.core_api()
    taint = kubernetes.client.V1Taint(
        key=NODE_HEALTH_LABEL,
        value='true',
        effect='NoSchedule',
    )
    node = core_api.read_node(
        name=node_name,
        _request_timeout=kube_client.API_TIMEOUT,
    )

    if node.spec.taints is None:
        node.spec.taints = []

    # duplicate taints are disallowed
    tainted = any(taint.key == NODE_HEALTH_LABEL for taint in node.spec.taints)
    if not tainted:
        node.spec.taints.append(taint)

    # Patch the node with the new taints
    core_api.patch_node(
        name=node_name,
        body=node,
        _request_timeout=kube_client.API_TIMEOUT,
    )

    logger.info(f'Node {node_name} tainted.')


def list_nodes() -> List[str]:
    """returns a list of k8s node names

    Returns:
        List[str]: List of k8s node names
    """
    nodes = kube_client.core_api().list_node()
    return [node.metadata.name for node in nodes.items]



# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: skypilot: https://github.com/skypilot-org/skypilot
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for subprocesses."""

import os
import resource
import subprocess
from multiprocessing import pool
from typing import Any, Callable, Dict, List, Optional, Union

import colorama
import psutil

from konduktor import constants, logging
from konduktor.utils import exceptions, log_utils, ux_utils

logger = logging.get_logger(__name__)

_fd_limit_warning_shown = False


def run(cmd, **kwargs):
    # Should be careful to use this function, as the child process cmd spawn may
    # keep running in the background after the current program is killed. To get
    # rid of this problem, use `log_utils.run_with_log`.
    shell = kwargs.pop('shell', True)
    check = kwargs.pop('check', True)
    executable = kwargs.pop('executable', '/bin/bash')
    if not shell:
        executable = None
    return subprocess.run(
        cmd, shell=shell, check=check, executable=executable, **kwargs
    )


def run_no_outputs(cmd, **kwargs):
    return run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, **kwargs)


def _get_thread_multiplier(cloud_str: Optional[str] = None) -> int:
    # If using Kubernetes, we use 4x the number of cores.
    if cloud_str and cloud_str.lower() == 'kubernetes':
        return 4
    return 1


def get_max_workers_for_file_mounts(
    common_file_mounts: Dict[str, str], cloud_str: Optional[str] = None
) -> int:
    global _fd_limit_warning_shown
    fd_limit, _ = resource.getrlimit(resource.RLIMIT_NOFILE)

    # Raise warning for low fd_limit (only once)
    if fd_limit < 1024 and not _fd_limit_warning_shown:
        logger.warning(
            f'Open file descriptor limit ({fd_limit}) is low. File sync to '
            'remote clusters may be slow. Consider increasing the limit using '
            '`ulimit -n <number>` or modifying system limits.'
        )
        _fd_limit_warning_shown = True

    fd_per_rsync = 5
    for src in common_file_mounts.values():
        if os.path.isdir(src):
            # Assume that each file/folder under src takes 5 file descriptors
            # on average.
            fd_per_rsync = max(fd_per_rsync, len(os.listdir(src)) * 5)

    # Reserve some file descriptors for the system and other processes
    fd_reserve = 100

    max_workers = (fd_limit - fd_reserve) // fd_per_rsync
    # At least 1 worker, and avoid too many workers overloading the system.
    num_threads = get_parallel_threads(cloud_str)
    max_workers = min(max(max_workers, 1), num_threads)
    logger.debug(f'Using {max_workers} workers for file mounts.')
    return max_workers


def get_parallel_threads(cloud_str: Optional[str] = None) -> int:
    """Returns the number of threads to use for parallel execution.

    Args:
        cloud_str: The cloud
    """
    cpu_count = os.cpu_count()
    if cpu_count is None:
        cpu_count = 1
    return max(4, cpu_count - 1) * _get_thread_multiplier(cloud_str)


def run_in_parallel(
    func: Callable, args: List[Any], num_threads: Optional[int] = None
) -> List[Any]:
    """Run a function in parallel on a list of arguments.

    Args:
        func: The function to run in parallel
        args: Iterable of arguments to pass to func
        num_threads: Number of threads to use. If None, uses
          get_parallel_threads()

    Returns:
      A list of the return values of the function func, in the same order as the
        arguments.

    Raises:
        Exception: The first exception encountered.
    """
    # Short-circuit for short lists
    if len(args) == 0:
        return []
    if len(args) == 1:
        return [func(args[0])]

    processes = num_threads if num_threads is not None else get_parallel_threads()

    with pool.ThreadPool(processes=processes) as p:
        ordered_iterators = p.imap(func, args)
        return list(ordered_iterators)


def handle_returncode(
    returncode: int,
    command: str,
    error_msg: Union[str, Callable[[], str]],
    stderr: Optional[str] = None,
    stream_logs: bool = True,
) -> None:
    """Handle the returncode of a command.

    Args:
        returncode: The returncode of the command.
        command: The command that was run.
        error_msg: The error message to print.
        stderr: The stderr of the command.
        stream_logs: Whether to stream logs.
    """
    echo = logger.error if stream_logs else logger.debug
    if returncode != 0:
        if stderr is not None:
            echo(stderr)

        if callable(error_msg):
            error_msg = error_msg()
        format_err_msg = f'{colorama.Fore.RED}{error_msg}{colorama.Style.RESET_ALL}'
        with ux_utils.print_exception_no_traceback():
            raise exceptions.CommandError(returncode, command, format_err_msg, stderr)


def kill_children_processes(
    parent_pids: Optional[Union[int, List[Optional[int]]]] = None, force: bool = False
) -> None:
    """Kill children processes recursively.

    We need to kill the children, so that
    1. The underlying subprocess will not print the logs to the terminal,
       after this program exits.
    2. The underlying subprocess will not continue with starting a cluster
       etc. while we are cleaning up the clusters.

    Args:
        parent_pids: Optional PIDs of a series of processes. The processes and
          their children will be killed.  If a list of PID is specified, it is
          killed by the order in the list. This is for guaranteeing the order
          of cleaning up and suppress flaky errors.
        force: bool, send SIGKILL if force, otherwise, use SIGTERM for
          gracefully kill the process.
    """
    if isinstance(parent_pids, int):
        parent_pids = [parent_pids]

    def kill(proc: psutil.Process):
        if not proc.is_running():
            # Skip if the process is not running.
            return
        logger.debug(f'Killing process {proc.pid}')
        try:
            if force:
                proc.kill()
            else:
                proc.terminate()
            proc.wait(timeout=10)
        except psutil.NoSuchProcess:
            # The child process may have already been terminated.
            pass
        except psutil.TimeoutExpired:
            logger.debug(f'Process {proc.pid} did not terminate after 10 seconds')
            # Attempt to force kill if the normal termination fails
            if not force:
                logger.debug(f'Force killing process {proc.pid}')
                proc.kill()
                proc.wait(timeout=5)  # Shorter timeout after force kill

    parent_processes = []
    if parent_pids is None:
        parent_processes = [psutil.Process()]
    else:
        for pid in parent_pids:
            try:
                process = psutil.Process(pid)
            except psutil.NoSuchProcess:
                continue
            parent_processes.append(process)

    for parent_process in parent_processes:
        child_processes = parent_process.children(recursive=True)
        if parent_pids is not None:
            kill(parent_process)
        logger.debug(f'Killing child processes: {child_processes}')
        for child in child_processes:
            kill(child)


def kill_process_daemon(process_pid: int) -> None:
    """Start a daemon as a safety net to kill the process.

    Args:
        process_pid: The PID of the process to kill.
    """
    # Get initial children list
    try:
        process = psutil.Process(process_pid)
        initial_children = [p.pid for p in process.children(recursive=True)]
    except psutil.NoSuchProcess:
        initial_children = []

    parent_pid = os.getpid()
    daemon_script = os.path.join(
        os.path.dirname(os.path.abspath(log_utils.__file__)), 'subprocess_daemon.py'
    )
    python_path = subprocess.check_output(
        constants.GET_PYTHON_PATH_CMD,
        shell=True,
        stderr=subprocess.DEVNULL,
        encoding='utf-8',
    ).strip()
    daemon_cmd = [
        python_path,
        daemon_script,
        '--parent-pid',
        str(parent_pid),
        '--proc-pid',
        str(process_pid),
        # We pass the initial children list to avoid the race condition where
        # the process_pid is terminated before the daemon starts and gets the
        # children list.
        '--initial-children',
        ','.join(map(str, initial_children)),
    ]

    # We do not need to set `start_new_session=True` here, as the
    # daemon script will detach itself from the parent process with
    # fork to avoid being killed by parent process. See the reason we
    # daemonize the process in `sky/skylet/subprocess_daemon.py`.
    subprocess.Popen(
        daemon_cmd,
        # Suppress output
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        # Disable input
        stdin=subprocess.DEVNULL,
    )



"""Loki utils: query/tail logs from Loki"""
# TODO(asaiacai): eventually support querying
# centralized loki that lives outside the cluster

import asyncio
import json
import urllib.parse

import colorama
import kr8s
import websockets

from konduktor import logging

logger = logging.get_logger(__name__)

LOKI_REMOTE_PORT = 3100
WEBSOCKET_TIMEOUT = 10
INFINITY = 999999


async def _read_loki_logs(loki_url: str, timeout: int, job_name: str, worker_id: int):
    ws = await asyncio.wait_for(websockets.connect(loki_url), timeout=WEBSOCKET_TIMEOUT)
    logger.info(
        f'{colorama.Fore.YELLOW}Tailing logs from Loki. '
        f'Forwarding from remote port {LOKI_REMOTE_PORT}. Press Ctrl+C to stop. '
        f'{colorama.Style.RESET_ALL}'
    )
    try:
        while True:
            message = await asyncio.wait_for(ws.recv(), timeout=timeout)
            try:
                payload = json.loads(message)
                for stream in payload['streams']:
                    if stream['values'][0][1] is not None:
                        print(
                            f"{colorama.Fore.CYAN}{colorama.Style.BRIGHT} "
                            f"(job_name={job_name} worker_id={worker_id})"
                            f"{colorama.Style.RESET_ALL} {stream['values'][0][1]}",
                            flush=True,
                        )
            except json.JSONDecodeError:
                logger.warning(f'Failed to decode log skipping: {message}')
                logger.debug(f'Dropped log: {message}')
                continue
    except asyncio.exceptions.TimeoutError:
        logger.debug('Websocket timed-out, closing log stream!')
    except KeyboardInterrupt:
        logger.debug('Keyboard interrupt, closing log stream!')


def tail_loki_logs_ws(
    job_name: str, worker_id: int = 0, num_logs: int = 1000, follow: bool = True
):
    if num_logs > 5000:
        # TODO(asaiacai): we should not have a limit on the number of logs, but rather
        # let the user specify any number of lines, and we can print the last N lines.
        # this can be done in chunks. Potentially, we can query range
        # until we reach the end of the log and then invoke tail again.
        # Also include checks that the job is running/ever ran.
        raise ValueError('num_logs must be less than or equal to 5000')
    loki_svc = kr8s.objects.Service.get('loki', namespace='loki')
    with kr8s.portforward.PortForward(
        loki_svc, LOKI_REMOTE_PORT, local_port='auto'
    ) as port:
        loki_url = f'ws://localhost:{port}/loki/api/v1/tail'
        logger.debug(f'Loki URL: {loki_url}')
        params = {
            'query': urllib.parse.quote(
                f'{{k8s_job_name="{job_name}-workers-0"}} '
                f' | batch_kubernetes_io_job_completion_index = `{worker_id}`'
            ),
            'limit': num_logs,
            'delay': 5,
            # TODO(asaiacai): need to auto-generate the start and end times.
        }

        query_string = '&'.join(f'{key}={value}' for key, value in params.items())
        loki_url += f'?{query_string}'
        timeout = INFINITY if follow else WEBSOCKET_TIMEOUT
        asyncio.run(_read_loki_logs(loki_url, timeout, job_name, worker_id))


# TODO(asaiacai): write a query_range function to get all the
# logs for a job for not tailing option

# Run the WebSocket log tailing function
if __name__ == '__main__':
    tail_loki_logs_ws('tune-bc43', worker_id=0, follow=True)



# The name for the environment variable that stores KONDUKTOR user hash
USER_ID_ENV_VAR = 'KONDUKTOR_USER_ID'

# The name for the environment variable that stores KONDUKTOR user name.
USER_ENV_VAR = 'KONDUKTOR_USER'



# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: skypilot: https://github.com/skypilot-org/skypilot
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import io
import multiprocessing
import os
import subprocess
import sys
import types
from typing import List, Optional, Tuple, Type, Union

import prettytable

from konduktor.utils import subprocess_utils


class LineProcessor(object):
    """A processor for log lines."""

    def __enter__(self) -> None:
        pass

    def process_line(self, log_line: str) -> None:
        pass

    def __exit__(
        self,
        except_type: Optional[Type[BaseException]],
        except_value: Optional[BaseException],
        traceback: Optional[types.TracebackType],
    ) -> None:
        del except_type, except_value, traceback  # unused
        pass


class _ProcessingArgs:
    """Arguments for processing logs."""

    def __init__(
        self,
        log_path: str,
        stream_logs: bool,
        start_streaming_at: str = '',
        end_streaming_at: Optional[str] = None,
        skip_lines: Optional[List[str]] = None,
        replace_crlf: bool = False,
        line_processor: Optional[LineProcessor] = None,
        streaming_prefix: Optional[str] = None,
    ) -> None:
        self.log_path = log_path
        self.stream_logs = stream_logs
        self.start_streaming_at = start_streaming_at
        self.end_streaming_at = end_streaming_at
        self.skip_lines = skip_lines
        self.replace_crlf = replace_crlf
        self.line_processor = line_processor
        self.streaming_prefix = streaming_prefix


def _handle_io_stream(io_stream, out_stream, args: _ProcessingArgs):
    """Process the stream of a process."""
    out_io = io.TextIOWrapper(
        io_stream, encoding='utf-8', newline='', errors='replace', write_through=True
    )

    start_streaming_flag = False
    end_streaming_flag = False
    streaming_prefix = args.streaming_prefix if args.streaming_prefix else ''
    line_processor = (
        LineProcessor() if args.line_processor is None else args.line_processor
    )

    out = []
    with open(args.log_path, 'a', encoding='utf-8') as fout:
        with line_processor:
            while True:
                line = out_io.readline()
                if not line:
                    break
                # start_streaming_at logic in processor.process_line(line)
                if args.replace_crlf and line.endswith('\r\n'):
                    # Replace CRLF with LF to avoid ray logging to the same
                    # line due to separating lines with '\n'.
                    line = line[:-2] + '\n'
                if args.skip_lines is not None and any(
                    skip in line for skip in args.skip_lines
                ):
                    continue
                if args.start_streaming_at in line:
                    start_streaming_flag = True
                if args.end_streaming_at is not None and args.end_streaming_at in line:
                    # Keep executing the loop, only stop streaming.
                    # E.g., this is used for `sky bench` to hide the
                    # redundant messages of `sky launch` while
                    # saving them in log files.
                    end_streaming_flag = True
                if args.stream_logs and start_streaming_flag and not end_streaming_flag:
                    print(streaming_prefix + line, end='', file=out_stream, flush=True)
                if args.log_path != '/dev/null':
                    fout.write(line)
                    fout.flush()
                line_processor.process_line(line)
                out.append(line)
    return ''.join(out)


def process_subprocess_stream(proc, args: _ProcessingArgs) -> Tuple[str, str]:
    """Redirect the process's filtered stdout/stderr to both stream and file"""
    if proc.stderr is not None:
        # Asyncio does not work as the output processing can be executed in a
        # different thread.
        # selectors is possible to handle the multiplexing of stdout/stderr,
        # but it introduces buffering making the output not streaming.
        with multiprocessing.pool.ThreadPool(processes=1) as pool:
            err_args = copy.copy(args)
            err_args.line_processor = None
            stderr_fut = pool.apply_async(
                _handle_io_stream, args=(proc.stderr, sys.stderr, err_args)
            )
            # Do not launch a thread for stdout as the rich.status does not
            # work in a thread, which is used in
            # log_utils.RayUpLineProcessor.
            stdout = _handle_io_stream(proc.stdout, sys.stdout, args)
            stderr = stderr_fut.get()
    else:
        stdout = _handle_io_stream(proc.stdout, sys.stdout, args)
        stderr = ''
    return stdout, stderr


def create_table(field_names: List[str], **kwargs) -> prettytable.PrettyTable:
    """Creates table with default style."""
    border = kwargs.pop('border', False)
    align = kwargs.pop('align', 'l')
    table = prettytable.PrettyTable(
        align=align, border=border, field_names=field_names, **kwargs
    )
    table.left_padding_width = 0
    table.right_padding_width = 2
    return table


def run_with_log(
    cmd: Union[List[str], str],
    log_path: str,
    *,
    require_outputs: bool = False,
    stream_logs: bool = False,
    start_streaming_at: str = '',
    end_streaming_at: Optional[str] = None,
    skip_lines: Optional[List[str]] = None,
    shell: bool = False,
    with_ray: bool = False,
    process_stream: bool = True,
    line_processor: Optional[LineProcessor] = None,
    streaming_prefix: Optional[str] = None,
    **kwargs,
) -> Union[int, Tuple[int, str, str]]:
    """Runs a command and logs its output to a file.

    Args:
        cmd: The command to run.
        log_path: The path to the log file.
        stream_logs: Whether to stream the logs to stdout/stderr.
        require_outputs: Whether to return the stdout/stderr of the command.
        process_stream: Whether to post-process the stdout/stderr of the
            command, such as replacing or skipping lines on the fly. If
            enabled, lines are printed only when '\r' or '\n' is found.

    Returns the returncode or returncode, stdout and stderr of the command.
      Note that the stdout and stderr is already decoded.
    """
    assert process_stream or not require_outputs, (
        process_stream,
        require_outputs,
        'require_outputs should be False when process_stream is False',
    )

    log_path = os.path.expanduser(log_path)
    dirname = os.path.dirname(log_path)
    os.makedirs(dirname, exist_ok=True)
    # Redirect stderr to stdout when using ray, to preserve the order of
    # stdout and stderr.
    stdout_arg = stderr_arg = None
    if process_stream:
        stdout_arg = subprocess.PIPE
        stderr_arg = subprocess.PIPE if not with_ray else subprocess.STDOUT
    # Use stdin=subprocess.DEVNULL by default, as allowing inputs will mess up
    # the terminal output when typing in the terminal that starts the API
    # server.
    stdin = kwargs.pop('stdin', subprocess.DEVNULL)
    with subprocess.Popen(
        cmd,
        stdout=stdout_arg,
        stderr=stderr_arg,
        start_new_session=True,
        shell=shell,
        stdin=stdin,
        **kwargs,
    ) as proc:
        try:
            subprocess_utils.kill_process_daemon(proc.pid)
            stdout = ''
            stderr = ''

            if process_stream:
                if skip_lines is None:
                    skip_lines = []
                # Skip these lines caused by `-i` option of bash. Failed to
                # find other way to turn off these two warning.
                # https://stackoverflow.com/questions/13300764/how-to-tell-bash-not-to-issue-warnings-cannot-set-terminal-process-group-and # noqa: E501
                # `ssh -T -i -tt` still cause the problem.
                skip_lines += [
                    'bash: cannot set terminal process group',
                    'bash: no job control in this shell',
                ]
                # We need this even if the log_path is '/dev/null' to ensure the
                # progress bar is shown.
                # NOTE: Lines are printed only when '\r' or '\n' is found.
                args = _ProcessingArgs(
                    log_path=log_path,
                    stream_logs=stream_logs,
                    start_streaming_at=start_streaming_at,
                    end_streaming_at=end_streaming_at,
                    skip_lines=skip_lines,
                    line_processor=line_processor,
                    # Replace CRLF when the output is logged to driver by ray.
                    replace_crlf=with_ray,
                    streaming_prefix=streaming_prefix,
                )
                stdout, stderr = process_subprocess_stream(proc, args)
            proc.wait()
            if require_outputs:
                return proc.returncode, stdout, stderr
            return proc.returncode
        except KeyboardInterrupt:
            # Kill the subprocess directly, otherwise, the underlying
            # process will only be killed after the python program exits,
            # causing the stream handling stuck at `readline`.
            subprocess_utils.kill_children_processes()
            raise



# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: skypilot: https://github.com/skypilot-org/skypilot
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import urllib.parse
from multiprocessing import pool
from typing import Any, Callable, Dict, List, Optional, Tuple

from konduktor import logging
from konduktor.adaptors import aws, gcp
from konduktor.utils import exceptions, log_utils, ux_utils

Client = Any

logger = logging.get_logger(__name__)


def split_gcs_path(gcs_path: str) -> Tuple[str, str]:
    """Splits GCS Path into Bucket name and Relative Path to Bucket

    Args:
      gcs_path: str; GCS Path, e.g. gcs://imagenet/train/
    """
    path_parts = gcs_path.replace('gs://', '').split('/')
    bucket = path_parts.pop(0)
    key = '/'.join(path_parts)
    return bucket, key


def split_s3_path(s3_path: str) -> Tuple[str, str]:
    """Splits S3 Path into Bucket name and Relative Path to Bucket

    Args:
      s3_path: str; S3 Path, e.g. s3://imagenet/train/
    """
    path_parts = s3_path.replace('s3://', '').split('/')
    bucket = path_parts.pop(0)
    key = '/'.join(path_parts)
    return bucket, key


def create_s3_client(region: Optional[str] = None) -> Client:
    """Helper method that connects to Boto3 client for S3 Bucket

    Args:
      region: str; Region name, e.g. us-west-1, us-east-2. If None, default
        region us-east-1 is used.
    """
    if region is None:
        region = 'us-east-1'
    return aws.client('s3', region_name=region)


def verify_gcs_bucket(name: str) -> bool:
    """Helper method that checks if the GCS bucket exists

    Args:
      name: str; Name of GCS Bucket (without gs:// prefix)
    """
    try:
        gcp.storage_client().get_bucket(name)
        return True
    except gcp.not_found_exception():
        return False


def verify_s3_bucket(name: str) -> bool:
    """Helper method that checks if the S3 bucket exists

    Args:
      name: str; Name of the S3 Bucket (without s3:// prefix)
    """
    import boto3
    from botocore.exceptions import ClientError

    s3 = boto3.client('s3')
    try:
        s3.head_bucket(Bucket=name)
        return True
    except ClientError as e:
        error_code = int(e.response['Error']['Code'])
        if error_code == 404:
            return False
        # In case of permissions issues or other errors, you can log and assume False
        return False


def is_cloud_store_url(url):
    result = urllib.parse.urlsplit(url)
    # '' means non-cloud URLs.
    return result.netloc


def _group_files_by_dir(
    source_list: List[str],
) -> Tuple[Dict[str, List[str]], List[str]]:
    """Groups a list of paths based on their directory

    Given a list of paths, generates a dict of {dir_name: List[file_name]}
    which groups files with same dir, and a list of dirs in the source_list.

    This is used to optimize uploads by reducing the number of calls to rsync.
    E.g., ['a/b/c.txt', 'a/b/d.txt', 'a/e.txt'] will be grouped into
    {'a/b': ['c.txt', 'd.txt'], 'a': ['e.txt']}, and these three files can be
    uploaded in two rsync calls instead of three.

    Args:
        source_list: List[str]; List of paths to group
    """
    grouped_files: Dict[str, List[str]] = {}
    dirs = []
    for source in source_list:
        source = os.path.abspath(os.path.expanduser(source))
        if os.path.isdir(source):
            dirs.append(source)
        else:
            base_path = os.path.dirname(source)
            file_name = os.path.basename(source)
            if base_path not in grouped_files:
                grouped_files[base_path] = []
            grouped_files[base_path].append(file_name)
    return grouped_files, dirs


def parallel_upload(
    source_path_list: List[str],
    filesync_command_generator: Callable[[str, List[str]], str],
    dirsync_command_generator: Callable[[str, str], str],
    log_path: str,
    bucket_name: str,
    access_denied_message: str,
    create_dirs: bool = False,
    max_concurrent_uploads: Optional[int] = None,
) -> None:
    """Helper function to run parallel uploads for a list of paths.

    Used by Store to run rsync commands in parallel by
    providing appropriate command generators.

    Args:
        source_path_list: List of paths to local files or directories
        filesync_command_generator: Callable that generates rsync command
            for a list of files belonging to the same dir.
        dirsync_command_generator: Callable that generates rsync command
            for a directory.
        log_path: Path to the log file
        bucket_name: Name of the bucket
        access_denied_message: Message to intercept from the underlying
            upload utility when permissions are insufficient. Used in
            exception handling.
        create_dirs: If the local_path is a directory and this is set to
            False, the contents of the directory are directly uploaded to
            root of the bucket. If the local_path is a directory and this is
            set to True, the directory is created in the bucket root and
            contents are uploaded to it.
        max_concurrent_uploads: Maximum number of concurrent threads to use
            to upload files.
    """
    # Generate gsutil rsync command for files and dirs
    commands = []
    grouped_files, dirs = _group_files_by_dir(source_path_list)
    # Generate file upload commands
    for dir_path, file_names in grouped_files.items():
        sync_command = filesync_command_generator(dir_path, file_names)
        commands.append(sync_command)
    # Generate dir upload commands
    for dir_path in dirs:
        if create_dirs:
            dest_dir_name = os.path.basename(dir_path)
        else:
            dest_dir_name = ''
        sync_command = dirsync_command_generator(dir_path, dest_dir_name)
        commands.append(sync_command)

    # Run commands in parallel
    with pool.ThreadPool(processes=max_concurrent_uploads) as p:
        p.starmap(
            run_upload_cli,
            zip(
                commands,
                [access_denied_message] * len(commands),
                [bucket_name] * len(commands),
                [log_path] * len(commands),
            ),
        )


def get_gsutil_command() -> Tuple[str, str]:
    """Gets the alias'd command for gsutil and a command to define the alias.

    This is required for applying platform-specific flags to gsutil.

    In particular, we disable multiprocessing on Mac using
    `-o "GSUtil:parallel_process_count=1"`. Multithreading is still enabled.
    gsutil on Mac has a bug with multiprocessing that causes it to crash
    when uploading files. Related issues:
    https://bugs.python.org/issue33725
    https://github.com/GoogleCloudPlatform/gsutil/issues/464

    The flags are added by checking the platform using bash in a one-liner.
    The platform check is done inline to have the flags match where the command
    is executed, rather than where the code is run. This is important when
    the command is run in a remote VM.

    Returns:
        Tuple[str, str] : (gsutil_alias, command to generate the alias)
        The command to generate alias must be run before using the alias. E.g.,
        ```
        gsutil_alias, alias_gen = get_gsutil_command()
        cmd_to_run = f'{alias_gen}; {gsutil_alias} cp ...'
        ```
    """
    gsutil_alias = 'skypilot_gsutil'
    disable_multiprocessing_flag = '-o "GSUtil:parallel_process_count=1"'

    # Define skypilot_gsutil as a shell function instead of an alias.
    # This function will behave just like alias, but can be called immediately
    # after its definition on the same line
    alias_gen = (
        f'[[ "$(uname)" == "Darwin" ]] && {gsutil_alias}() {{ '
        f'gsutil -m {disable_multiprocessing_flag} "$@"; }} '
        f'|| {gsutil_alias}() {{ gsutil -m "$@"; }}'
    )

    return gsutil_alias, alias_gen


def run_upload_cli(
    command: str, access_denied_message: str, bucket_name: str, log_path: str
):
    returncode, stdout, stderr = log_utils.run_with_log(  # type: ignore[misc]
        command,
        log_path,
        shell=True,
        require_outputs=True,
        # We need to use bash as some of the cloud commands uses bash syntax,
        # such as [[ ... ]]
        executable='/bin/bash',
    )
    if access_denied_message in stderr:
        with ux_utils.print_exception_no_traceback():
            raise PermissionError(
                'Failed to upload files to '
                'the remote bucket. The bucket does not have '
                'write permissions. It is possible that '
                'the bucket is public.'
            )
    if returncode != 0:
        with ux_utils.print_exception_no_traceback():
            logger.error(stderr)
            raise exceptions.StorageUploadError(
                f'Upload to bucket failed for store {bucket_name}. '
                f'Please check the logs: {log_path}'
            )
    if not stdout:
        logger.debug(
            'No file uploaded. This could be due to an error or '
            'because all files already exist on the cloud.'
        )



# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: skypilot: https://github.com/skypilot-org/skypilot
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Konduktor backend interface."""

import typing
from typing import Dict, Optional

if typing.TYPE_CHECKING:
    from konduktor.data import storage as storage_lib

import konduktor
from konduktor.utils import ux_utils

Path = str


class Backend:
    """Backend interface: handles provisioning, setup, and scheduling."""

    # NAME is used to identify the backend class from cli/yaml.
    NAME = 'backend'

    # --- APIs ---
    def check_resources_fit_cluster(self, task: 'konduktor.Task') -> bool:
        """Check whether resources of the task are satisfied by cluster."""
        raise NotImplementedError

    def sync_workdir(self, workdir: Path) -> None:
        return self._sync_workdir(workdir)

    def sync_file_mounts(
        self,
        all_file_mounts: Optional[Dict[Path, Path]],
        storage_mounts: Optional[Dict[Path, 'storage_lib.Storage']],
    ) -> None:
        return self._sync_file_mounts(all_file_mounts, storage_mounts)

    def add_storage_objects(self, task: 'konduktor.Task') -> None:
        raise NotImplementedError

    def execute(
        self, task: 'konduktor.Task', detach_run: bool, dryrun: bool = False
    ) -> Optional[str]:
        """Execute the task on the cluster.

        Returns:
            Job id if the task is submitted to the cluster, None otherwise.
        """
        ux_utils.spinner_message('Submitting job')
        return self._execute(task, detach_run, dryrun)

    def post_execute(self) -> None:
        """Post execute(): e.g., print helpful inspection messages."""
        return self._post_execute()

    def register_info(self, **kwargs) -> None:
        """Register backend-specific information."""
        pass

    def _sync_workdir(self, workdir: Path) -> None:
        raise NotImplementedError

    def _sync_file_mounts(
        self,
        all_file_mounts: Optional[Dict[Path, Path]],
        storage_mounts: Optional[Dict[Path, 'storage_lib.Storage']],
    ) -> None:
        raise NotImplementedError

    def _execute(
        self, task: 'konduktor.Task', detach_run: bool, dryrun: bool = False
    ) -> Optional[str]:
        raise NotImplementedError

    def _post_execute(self) -> None:
        raise NotImplementedError



"""Jobset utils: wraps CRUD operations for jobsets"""

import enum
import json
import os
import tempfile
import typing
from datetime import datetime, timezone
from typing import Any, Dict, Optional
from urllib.parse import urlparse

import colorama

import konduktor
from konduktor import constants, kube_client, logging
from konduktor.data import registry
from konduktor.utils import common_utils, kubernetes_utils, log_utils

if typing.TYPE_CHECKING:
    pass

logger = logging.get_logger(__name__)

JOBSET_API_GROUP = 'jobset.x-k8s.io'
JOBSET_API_VERSION = 'v1alpha2'
JOBSET_PLURAL = 'jobsets'

JOBSET_NAME_LABEL = 'trainy.ai/job-name'
JOBSET_USERID_LABEL = 'trainy.ai/user-id'
JOBSET_USER_LABEL = 'trainy.ai/username'
JOBSET_ACCELERATOR_LABEL = 'trainy.ai/accelerator'
JOBSET_NUM_ACCELERATORS_LABEL = 'trainy.ai/num-accelerators'


_JOBSET_METADATA_LABELS = {
    'jobset_name_label': JOBSET_NAME_LABEL,
    'jobset_userid_label': JOBSET_USERID_LABEL,
    'jobset_user_label': JOBSET_USER_LABEL,
    'jobset_accelerator_label': JOBSET_ACCELERATOR_LABEL,
    'jobset_num_accelerators_label': JOBSET_NUM_ACCELERATORS_LABEL,
}
_RUN_DURATION_ANNOTATION = 'maxRunDurationSeconds'
_RUN_DURATION_ANNOTATION_KEY = 'kueue.x-k8s.io/maxRunDurationSeconds'


class JobNotFoundError(Exception):
    pass


class JobStatus(enum.Enum):
    SUSPENDED = 'SUSPENDED'
    ACTIVE = 'ACTIVE'
    COMPLETED = 'COMPLETED'
    FAILED = 'FAILED'
    PENDING = 'PENDING'


if typing.TYPE_CHECKING:
    import konduktor


def create_pod_spec(task: 'konduktor.Task') -> Dict[str, Any]:
    """Merges the task defintion with config
    to create a final pod spec dict for the job

    Returns:
        Dict[str, Any]: k8s pod spec
    """

    # fill out the templating variables
    assert task.resources is not None, 'Task resources are required'
    if task.resources.accelerators:
        num_gpus = list(task.resources.accelerators.values())[0]
    else:
        num_gpus = 0
    task.name = f'{task.name}-{common_utils.get_usage_run_id()[:4]}'
    node_hostnames = ','.join(
        [f'{task.name}-workers-0-{idx}.{task.name}' for idx in range(task.num_nodes)]
    )
    master_addr = f'{task.name}-workers-0-0.{task.name}'

    if task.resources.accelerators:
        accelerator_type = list(task.resources.accelerators.keys())[0]
    else:
        accelerator_type = None

    # template the commands to run on the container for syncing files. At this point
    # task.stores is Dict[str, storage_utils.Storage] which is (dst, storage_obj_src)
    # first we iterate through storage_mounts and then file_mounts.
    sync_commands = []
    mkdir_commands = []
    storage_secrets = {}

    # first do storage_mount sync
    for dst, store in task.storage_mounts.items():
        # TODO(asaiacai) idk why but theres an extra storage mount for the
        # file mounts. Should be cleaned up eventually in
        # maybe_translate_local_file_mounts_and_sync_up
        assert store.source is not None and isinstance(
            store.source, str
        ), 'Store source is required'
        store_scheme = urlparse(store.source).scheme
        if '/tmp/konduktor-job-filemounts-files' in dst:
            continue
        # should impelement a method here instead of raw dog dict access
        cloud_store = registry._REGISTRY[store_scheme]
        storage_secrets[store_scheme] = cloud_store._STORE.get_k8s_credential_name()
        mkdir_commands.append(
            f'cd {constants.KONDUKTOR_REMOTE_WORKDIR};' f'mkdir -p {dst}'
        )
        assert store._bucket_sub_path is not None
        sync_commands.append(
            cloud_store.make_sync_dir_command(
                os.path.join(store.source, store._bucket_sub_path), dst
            )
        )

    # then do file_mount sync.
    assert task.file_mounts is not None
    for dst, src in task.file_mounts.items():
        store_scheme = str(urlparse(store.source).scheme)
        cloud_store = registry._REGISTRY[store_scheme]
        mkdir_commands.append(
            f'cd {constants.KONDUKTOR_REMOTE_WORKDIR};'
            f'mkdir -p {os.path.dirname(dst)}'
        )
        storage_secrets[store_scheme] = cloud_store._STORE.get_k8s_credential_name()
        sync_commands.append(cloud_store.make_sync_file_command(src, dst))

    assert task.resources is not None, 'Task resources are required'
    assert task.resources.cpus is not None, 'Task resources cpus are required'
    assert task.resources.memory is not None, 'Task resources memory are required'
    assert task.resources.image_id is not None, 'Task resources image_id are required'
    with tempfile.NamedTemporaryFile() as temp:
        common_utils.fill_template(
            'pod.yaml.j2',
            {
                # TODO(asaiacai) need to parse/round these numbers and sanity check
                'cpu': kubernetes_utils.parse_cpu_or_gpu_resource(task.resources.cpus),
                'memory': kubernetes_utils.parse_memory_resource(task.resources.memory),
                'image_id': task.resources.image_id,
                'num_gpus': num_gpus,
                'master_addr': master_addr,
                'num_nodes': task.num_nodes,
                'job_name': task.name,  # append timestamp and user id here?
                'run_cmd': task.run,
                'node_hostnames': node_hostnames,
                'accelerator_type': accelerator_type,
                'sync_commands': sync_commands,
                'mkdir_commands': mkdir_commands,
                'mount_secrets': storage_secrets,
                'remote_workdir': constants.KONDUKTOR_REMOTE_WORKDIR,
                'user': common_utils.get_cleaned_username(),
            },
            temp.name,
        )
        pod_config = common_utils.read_yaml(temp.name)
        # merge with `~/.konduktor/config.yaml``
        kubernetes_utils.combine_pod_config_fields(temp.name, pod_config)
        pod_config = common_utils.read_yaml(temp.name)
        for k, v in task.envs.items():
            pod_config['kubernetes']['pod_config']['spec']['containers'][0][
                'env'
            ].append({'name': k, 'value': v})

    # TODO(asaiacai): have some schema validations. see
    # https://github.com/skypilot-org/skypilot/pull/4466
    # TODO(asaiacai): where can we include policies for the pod spec.

    return pod_config


def create_jobset(
    namespace: str,
    task: 'konduktor.Task',
    pod_spec: Dict[str, Any],
    dryrun: bool = False,
) -> Optional[Dict[str, Any]]:
    """Creates a jobset based on the task definition and pod spec
    and returns the created jobset spec
    """
    assert task.resources is not None, 'Task resources are undefined'
    if task.resources.accelerators:
        accelerator_type = list(task.resources.accelerators.keys())[0]
        num_accelerators = list(task.resources.accelerators.values())[0]
    else:
        accelerator_type = 'None'
        num_accelerators = 0
    with tempfile.NamedTemporaryFile() as temp:
        common_utils.fill_template(
            'jobset.yaml.j2',
            {
                'job_name': task.name,
                'user_id': common_utils.user_and_hostname_hash(),
                'num_nodes': task.num_nodes,
                'user': common_utils.get_cleaned_username(),
                'accelerator_type': accelerator_type,
                'num_accelerators': num_accelerators,
                **_JOBSET_METADATA_LABELS,
            },
            temp.name,
        )
        jobset_spec = common_utils.read_yaml(temp.name)
        jobset_spec['jobset']['metadata']['labels'].update(**task.resources.labels)
        assert task.resources.labels is not None
        maxRunDurationSeconds = task.resources.labels.get('maxRunDurationSeconds', None)
        if not maxRunDurationSeconds:
            raise ValueError('maxRunDurationSeconds is required')
        jobset_spec['jobset']['metadata']['annotations'][
            _RUN_DURATION_ANNOTATION_KEY
        ] = str(maxRunDurationSeconds)
    jobset_spec['jobset']['spec']['replicatedJobs'][0]['template']['spec'][
        'template'
    ] = pod_spec  # noqa: E501
    try:
        jobset = kube_client.crd_api().create_namespaced_custom_object(
            group=JOBSET_API_GROUP,
            version=JOBSET_API_VERSION,
            namespace=namespace,
            plural=JOBSET_PLURAL,
            body=jobset_spec['jobset'],
            dry_run='All' if dryrun else None,
        )
        logger.info(
            f'task {colorama.Fore.CYAN}{colorama.Style.BRIGHT}'
            f'{task.name}{colorama.Style.RESET_ALL} created'
        )
        return jobset
    except kube_client.api_exception() as err:
        try:
            error_body = json.loads(err.body)
            error_message = error_body.get('message', '')
            logger.error(f'error creating jobset: {error_message}')
        except json.JSONDecodeError:
            error_message = str(err.body)
            logger.error(f'error creating jobset: {error_message}')
        else:
            # Re-raise the exception if it's a different error
            raise err
    return None


def list_jobset(namespace: str) -> Optional[Dict[str, Any]]:
    """Lists all jobsets in this namespace"""
    try:
        response = kube_client.crd_api().list_namespaced_custom_object(
            group=JOBSET_API_GROUP,
            version=JOBSET_API_VERSION,
            namespace=namespace,
            plural=JOBSET_PLURAL,
        )
        return response
    except kube_client.api_exception() as err:
        try:
            error_body = json.loads(err.body)
            error_message = error_body.get('message', '')
            logger.error(f'error listing jobset: {error_message}')
        except json.JSONDecodeError:
            error_message = str(err.body)
            logger.error(f'error creating jobset: {error_message}')
        else:
            # Re-raise the exception if it's a different error
            raise err
    return None


def get_jobset(namespace: str, job_name: str) -> Optional[Dict[str, Any]]:
    """Retrieves jobset in this namespace"""
    try:
        response = kube_client.crd_api().get_namespaced_custom_object(
            group=JOBSET_API_GROUP,
            version=JOBSET_API_VERSION,
            namespace=namespace,
            plural=JOBSET_PLURAL,
            name=job_name,
        )
        return response
    except kube_client.api_exception() as err:
        if err.status == 404:
            raise JobNotFoundError(
                f"Jobset '{job_name}' " f"not found in namespace '{namespace}'."
            )
        try:
            error_body = json.loads(err.body)
            error_message = error_body.get('message', '')
            logger.error(f'error getting jobset: {error_message}')
        except json.JSONDecodeError:
            error_message = str(err.body)
            logger.error(f'error creating jobset: {error_message}')
        else:
            # Re-raise the exception if it's a different error
            raise err
    return None


def delete_jobset(namespace: str, job_name: str) -> Optional[Dict[str, Any]]:
    """Deletes jobset in this namespace

    Args:
        namespace: Namespace where jobset exists
        job_name: Name of jobset to delete

    Returns:
        Response from delete operation
    """
    try:
        response = kube_client.crd_api().delete_namespaced_custom_object(
            group=JOBSET_API_GROUP,
            version=JOBSET_API_VERSION,
            namespace=namespace,
            plural=JOBSET_PLURAL,
            name=job_name,
        )
        return response
    except kube_client.api_exception() as err:
        try:
            error_body = json.loads(err.body)
            error_message = error_body.get('message', '')
            logger.error(f'error deleting jobset: {error_message}')
        except json.JSONDecodeError:
            error_message = str(err.body)
            logger.error(f'error deleting jobset: {error_message}')
        else:
            # Re-raise the exception if it's a different error
            raise err
    return None


def get_job(namespace: str, job_name: str) -> Optional[Dict[str, Any]]:
    """Gets a specific job from a jobset by name and worker index

    Args:
        namespace: Namespace where job exists
        job_name: Name of jobset containing the job
        worker_id: Index of the worker job to get (defaults to 0)

    Returns:
        Job object if found
    """
    try:
        # Get the job object using the job name
        # pattern {jobset-name}-workers-0-{worker_id}
        job_name = f'{job_name}-workers-0'
        response = kube_client.batch_api().read_namespaced_job(
            name=job_name, namespace=namespace
        )
        return response
    except kube_client.api_exception() as err:
        try:
            error_body = json.loads(err.body)
            error_message = error_body.get('message', '')
            logger.error(f'error getting job: {error_message}')
        except json.JSONDecodeError:
            error_message = str(err.body)
            logger.error(f'error getting job: {error_message}')
        else:
            # Re-raise the exception if it's a different error
            raise err
    return None


def show_status_table(namespace: str, all_users: bool):
    """Compute cluster table values and display.

    Returns:
        Number of pending auto{stop,down} clusters that are not already
        STOPPED.
    """
    # TODO(zhwu): Update the information for autostop clusters.

    def _get_status_string_colorized(status: Dict[str, Any]) -> str:
        terminalState = status.get('terminalState', None)
        if terminalState and terminalState.upper() == JobStatus.COMPLETED.name.upper():
            return (
                f'{colorama.Fore.GREEN}'
                f'{JobStatus.COMPLETED.name}{colorama.Style.RESET_ALL}'
            )
        elif terminalState and terminalState.upper() == JobStatus.FAILED.name.upper():
            return (
                f'{colorama.Fore.RED}'
                f'{JobStatus.FAILED.name}{colorama.Style.RESET_ALL}'
            )
        elif status['replicatedJobsStatus'][0]['ready']:
            return (
                f'{colorama.Fore.CYAN}'
                f'{JobStatus.ACTIVE.name}{colorama.Style.RESET_ALL}'
            )
        elif status['replicatedJobsStatus'][0]['suspended']:
            return (
                f'{colorama.Fore.GREEN}'
                f'{JobStatus.SUSPENDED.name}{colorama.Style.RESET_ALL}'
            )
        else:
            return (
                f'{colorama.Fore.BLUE}'
                f'{JobStatus.PENDING.name}{colorama.Style.RESET_ALL}'
            )

    def _get_time_delta(timestamp: str):
        delta = datetime.now(timezone.utc) - datetime.strptime(
            timestamp, '%Y-%m-%dT%H:%M:%SZ'
        ).replace(tzinfo=timezone.utc)
        total_seconds = int(delta.total_seconds())

        days, remainder = divmod(total_seconds, 86400)  # 86400 seconds in a day
        hours, remainder = divmod(remainder, 3600)  # 3600 seconds in an hour
        minutes, _ = divmod(remainder, 60)  # 60 seconds in a minute

        days_str = f'{days} days, ' if days > 0 else ''
        hours_str = f'{hours} hours, ' if hours > 0 else ''
        minutes_str = f'{minutes} minutes' if minutes > 0 else ''

        return f'{days_str}{hours_str}{minutes_str}'

    def _get_resources(job: Dict[str, Any]) -> str:
        num_pods = int(
            job['spec']['replicatedJobs'][0]['template']['spec']['parallelism']
        )  # noqa: E501
        resources = job['spec']['replicatedJobs'][0]['template']['spec']['template'][
            'spec'
        ]['containers'][0]['resources']['limits']  # noqa: E501
        cpu, memory = resources['cpu'], resources['memory']
        accelerator = job['metadata']['labels'].get(JOBSET_ACCELERATOR_LABEL, None)
        if accelerator:
            return f'{num_pods}x({cpu}CPU, memory {memory}, {accelerator})'
        else:
            return f'{num_pods}x({cpu}CPU, memory {memory}GB)'

    if all_users:
        columns = ['NAME', 'USER', 'STATUS', 'RESOURCES', 'SUBMITTED']
    else:
        columns = ['NAME', 'STATUS', 'RESOURCES', 'SUBMITTED']
    job_table = log_utils.create_table(columns)
    job_specs = list_jobset(namespace)
    assert job_specs is not None, 'Retrieving jobs failed'
    for job in job_specs['items']:
        if all_users:
            job_table.add_row(
                [
                    job['metadata']['name'],
                    job['metadata']['labels'][JOBSET_USERID_LABEL],
                    _get_status_string_colorized(job['status']),
                    _get_resources(job),
                    _get_time_delta(job['metadata']['creationTimestamp']),
                ]
            )
        elif (
            not all_users
            and job['metadata']['labels'][JOBSET_USER_LABEL]
            == common_utils.get_cleaned_username()
        ):
            job_table.add_row(
                [
                    job['metadata']['name'],
                    _get_status_string_colorized(job['status']),
                    _get_resources(job),
                    _get_time_delta(job['metadata']['creationTimestamp']),
                ]
            )
    print(job_table)



# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: skypilot: https://github.com/skypilot-org/skypilot
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for the storage module."""

import glob
import os
import shlex
import subprocess
import typing
from typing import List, Optional

import colorama

if typing.TYPE_CHECKING:
    from konduktor.data.constants import SourceType, StorageHandle


from konduktor import constants, logging
from konduktor.utils import common_utils, exceptions

logger = logging.get_logger(__name__)

_FILE_EXCLUSION_FROM_GITIGNORE_FAILURE_MSG = (
    f'{colorama.Fore.YELLOW}Warning: Files/dirs '
    'specified in .gitignore will be uploaded '
    'to the cloud storage for {path!r}'
    'due to the following error: {error_msg!r}'
)


def get_excluded_files_from_konduktorignore(src_dir_path: str) -> List[str]:
    """List files and patterns ignored by the .konduktor file
    in the given source directory.
    """
    excluded_list: List[str] = []
    expand_src_dir_path = os.path.expanduser(src_dir_path)
    konduktorignore_path = os.path.join(
        expand_src_dir_path, constants.KONDUKTOR_IGNORE_FILE
    )

    try:
        with open(konduktorignore_path, 'r', encoding='utf-8') as f:
            for line in f:
                line = line.strip()
                if line and not line.startswith('#'):
                    # Make parsing consistent with rsync.
                    # Rsync uses '/' as current directory.
                    if line.startswith('/'):
                        line = '.' + line
                    else:
                        line = '**/' + line
                    # Find all files matching the pattern.
                    matching_files = glob.glob(
                        os.path.join(expand_src_dir_path, line), recursive=True
                    )
                    # Process filenames to comply with cloud rsync format.
                    for i in range(len(matching_files)):
                        matching_files[i] = os.path.relpath(
                            matching_files[i], expand_src_dir_path
                        )
                    excluded_list.extend(matching_files)
    except IOError as e:
        logger.warning(
            f'Error reading {konduktorignore_path}: '
            f'{common_utils.format_exception(e, use_bracket=True)}'
        )

    return excluded_list


def get_excluded_files_from_gitignore(src_dir_path: str) -> List[str]:
    """Lists files and patterns ignored by git in the source directory

    Runs `git status --ignored` which returns a list of excluded files and
    patterns read from .gitignore and .git/info/exclude using git.
    `git init` is run if SRC_DIR_PATH is not a git repository and removed
    after obtaining excluded list.

    Returns:
        List[str] containing files and patterns to be ignored.  Some of the
        patterns include, **/mydir/*.txt, !myfile.log, or file-*/.
    """
    expand_src_dir_path = os.path.expanduser(src_dir_path)

    git_exclude_path = os.path.join(expand_src_dir_path, '.git/info/exclude')
    gitignore_path = os.path.join(expand_src_dir_path, constants.GIT_IGNORE_FILE)

    git_exclude_exists = os.path.isfile(git_exclude_path)
    gitignore_exists = os.path.isfile(gitignore_path)

    # This command outputs a list to be excluded according to .gitignore
    # and .git/info/exclude
    filter_cmd = (
        f'git -C {shlex.quote(expand_src_dir_path)} ' 'status --ignored --porcelain=v1'
    )
    excluded_list: List[str] = []

    if git_exclude_exists or gitignore_exists:
        try:
            output = subprocess.run(
                filter_cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
                text=True,
            )
        except subprocess.CalledProcessError as e:
            # when the SRC_DIR_PATH is not a git repo and .git
            # does not exist in it
            if e.returncode == exceptions.GIT_FATAL_EXIT_CODE:
                if 'not a git repository' in e.stderr:
                    # Check if the user has 'write' permission to
                    # SRC_DIR_PATH
                    if not os.access(expand_src_dir_path, os.W_OK):
                        error_msg = 'Write permission denial'
                        logger.warning(
                            _FILE_EXCLUSION_FROM_GITIGNORE_FAILURE_MSG.format(
                                path=src_dir_path, error_msg=error_msg
                            )
                        )
                        return excluded_list
                    init_cmd = f'git -C {expand_src_dir_path} init'
                    try:
                        subprocess.run(
                            init_cmd,
                            shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            check=True,
                        )
                        output = subprocess.run(
                            filter_cmd,
                            shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            check=True,
                            text=True,
                        )
                    except subprocess.CalledProcessError as init_e:
                        logger.warning(
                            _FILE_EXCLUSION_FROM_GITIGNORE_FAILURE_MSG.format(
                                path=src_dir_path, error_msg=init_e.stderr
                            )
                        )
                        return excluded_list
                    if git_exclude_exists:
                        # removes all the files/dirs created with 'git init'
                        # under .git/ except .git/info/exclude
                        remove_files_cmd = (
                            f'find {expand_src_dir_path}'
                            f'/.git -path {git_exclude_path}'
                            ' -prune -o -type f -exec rm -f '
                            '{} +'
                        )
                        remove_dirs_cmd = (
                            f'find {expand_src_dir_path}'
                            f'/.git -path {git_exclude_path}'
                            ' -o -type d -empty -delete'
                        )
                        subprocess.run(
                            remove_files_cmd,
                            shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            check=True,
                        )
                        subprocess.run(
                            remove_dirs_cmd,
                            shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            check=True,
                        )

        output_list = output.stdout.split('\n')
        for line in output_list:
            # FILTER_CMD outputs items preceded by '!!'
            # to specify excluded files/dirs
            # e.g., '!! mydir/' or '!! mydir/myfile.txt'
            if line.startswith('!!'):
                to_be_excluded = line[3:]
                if line.endswith('/'):
                    # aws s3 sync and gsutil rsync require * to exclude
                    # files/dirs under the specified directory.
                    to_be_excluded += '*'
                excluded_list.append(to_be_excluded)
    return excluded_list


def get_excluded_files(src_dir_path: str) -> List[str]:
    # TODO: this could return a huge list of files,
    # should think of ways to optimize.
    """List files and directories to be excluded."""
    expand_src_dir_path = os.path.expanduser(src_dir_path)
    konduktorignore_path = os.path.join(
        expand_src_dir_path, constants.KONDUKTOR_IGNORE_FILE
    )
    if os.path.exists(konduktorignore_path):
        logger.info(
            f'  {colorama.Style.DIM}'
            f'Excluded files to sync to cluster based on '
            f'{constants.KONDUKTOR_IGNORE_FILE}.'
            f'{colorama.Style.RESET_ALL}'
        )
        return get_excluded_files_from_konduktorignore(src_dir_path)
    logger.info(
        f'  {colorama.Style.DIM}'
        f'Excluded files to sync to cluster based on '
        f'{constants.GIT_IGNORE_FILE}.'
        f'{colorama.Style.RESET_ALL}'
    )
    return get_excluded_files_from_gitignore(src_dir_path)


class AbstractStore:
    """AbstractStore abstracts away the different storage types exposed by
    different clouds.

    Storage objects are backed by AbstractStores, each representing a store
    present in a cloud.
    """

    class StoreMetadata:
        """A pickle-able representation of Store

        Allows store objects to be written to and reconstructed from
        global_user_state.
        """

        def __init__(
            self,
            *,
            name: str,
            source: Optional['SourceType'],
            region: Optional[str] = None,
            is_sky_managed: Optional[bool] = None,
            _bucket_sub_path: Optional[str] = None,
        ):
            self.name = name
            self.source = source
            self.region = region
            self.is_sky_managed = is_sky_managed
            self._bucket_sub_path = _bucket_sub_path

        def __repr__(self):
            return (
                f'StoreMetadata('
                f'\n\tname={self.name},'
                f'\n\tsource={self.source},'
                f'\n\tregion={self.region},'
                f'\n\tis_sky_managed={self.is_sky_managed},'
                f'\n\t_bucket_sub_path={self._bucket_sub_path}'
            )

    def __init__(
        self,
        name: str,
        source: Optional['SourceType'],
        region: Optional[str] = None,
        is_sky_managed: Optional[bool] = None,
        sync_on_reconstruction: Optional[bool] = True,
        _bucket_sub_path: Optional[str] = None,
    ):
        """Initialize AbstractStore

        Args:
            name: Store name
            source: Data source for the store
            region: Region to place the bucket in
            is_sky_managed: Whether the store is managed by Sky. If None, it
              must be populated by the implementing class during initialization.

        Raises:
            StorageBucketCreateError: If bucket creation fails
            StorageBucketGetError: If fetching existing bucket fails
            StorageInitError: If general initialization fails
        """
        self.name = name
        self.source = source
        self.region = region
        self.is_sky_managed = is_sky_managed
        self.sync_on_reconstruction = sync_on_reconstruction
        self._bucket_sub_path = _bucket_sub_path
        # Whether sky is responsible for the lifecycle of the Store.
        self._validate()
        self.initialize()

    @property
    def bucket_sub_path(self) -> Optional[str]:
        """Get the bucket_sub_path."""
        return self._bucket_sub_path

    @classmethod
    def from_metadata(cls, metadata: StoreMetadata, **override_args):
        """Create a Store from a StoreMetadata object.

        Used when reconstructing Storage and Store objects from
        global_user_state.
        """
        return cls(
            name=override_args.get('name', metadata.name),
            source=override_args.get('source', metadata.source),
            region=override_args.get('region', metadata.region),
        )

    def get_metadata(self) -> StoreMetadata:
        return self.StoreMetadata(
            name=self.name,
            source=self.source,
            region=self.region,
        )

    def initialize(self):
        """Initializes the Store object on the cloud.

        Initialization involves fetching bucket if exists, or creating it if
        it does not.

        Raises:
          StorageBucketCreateError: If bucket creation fails
          StorageBucketGetError: If fetching existing bucket fails
          StorageInitError: If general initialization fails.
        """
        pass

    def _validate(self) -> None:
        """Runs validation checks on class args"""
        pass

    def upload(self) -> None:
        """Uploads source to the store bucket

        Upload must be called by the Storage handler - it is not called on
        Store initialization.
        """
        raise NotImplementedError

    def delete(self) -> None:
        """Removes the Storage object from the cloud."""
        raise NotImplementedError

    def get_handle(self) -> 'StorageHandle':
        """Returns the storage handle for use by the execution backend to attach
        to VM/containers
        """
        raise NotImplementedError

    def download_remote_dir(self, local_path: str) -> None:
        """Downloads directory from remote bucket to the specified
        local_path

        Args:
          local_path: Local path on user's device
        """
        raise NotImplementedError

    def _download_file(self, remote_path: str, local_path: str) -> None:
        """Downloads file from remote to local on Store

        Args:
          remote_path: str; Remote file path on Store
          local_path: str; Local file path on user's device
        """
        raise NotImplementedError

    def mount_command(self, mount_path: str) -> str:
        """Returns the command to mount the Store to the specified mount_path.

        Includes the setup commands to install mounting tools.

        Args:
          mount_path: str; Mount path on remote server
        """
        raise NotImplementedError

    def __deepcopy__(self, memo):
        # S3 Client and GCS Client cannot be deep copied, hence the
        # original Store object is returned
        return self

    def _validate_existing_bucket(self):
        """Validates the storage fields for existing buckets."""
        # Check if 'source' is None, this is only allowed when Storage is in
        # either MOUNT mode or COPY mode with sky-managed storage.
        # Note: In COPY mode, a 'source' being None with non-sky-managed
        # storage is already handled as an error in _validate_storage_spec.
        if self.source is None:
            # Retrieve a handle associated with the storage name.
            # This handle links to sky managed storage if it exists.
            raise NotImplementedError("We don't handle empty sources for now")

    @classmethod
    def check_credentials(cls):
        """
        Check if the credentials stored on client are valid for the store.
        This function always runs after check_credentials_from_secret. If
        the credentials work, we create/update the secret on the cluster.
        """
        raise NotImplementedError

    @classmethod
    def set_secret_credentials(cls):
        """
        Set the k8s secret storing the credentials for the store.
        """
        raise NotImplementedError

    # TODO(zhwu): Make the return type immutable.
    @classmethod
    def get_user_identities(cls) -> Optional[List[List[str]]]:
        """(Advanced) Returns all available user identities of this cloud.

        The user "identity" is associated with each SkyPilot cluster they
        create. This is used in protecting cluster operations, such as
        provision, teardown and status refreshing, in a multi-identity
        scenario, where the same user/device can switch between different
        cloud identities. We check that the user identity matches before:
            - Provisioning/starting a cluster
            - Stopping/tearing down a cluster
            - Refreshing the status of a cluster

        Design choices:
          1. We allow the operations that can correctly work with a different
             user identity, as a user should have full control over all their
             clusters (no matter which identity it belongs to), e.g.,
             submitting jobs, viewing logs, auto-stopping, etc.
          2. A cloud implementation can optionally switch between different
             identities if required for cluster operations. In this case,
             the cloud implementation should return multiple identities
             as a list. E.g., our Kubernetes implementation can use multiple
             kubeconfig contexts to switch between different identities.

        The choice of what constitutes an identity is up to each cloud's
        implementation. In general, to suffice for the above purposes,
        ensure that different identities should imply different sets of
        resources are used when the user invoked each cloud's default
        CLI/API.

        An identity is a list of strings. The list is in the order of
        strictness, i.e., the first element is the most strict identity, and
        the last element is the least strict identity.
        When performing an identity check between the current active identity
        and the owner identity associated with a cluster, we compare the two
        lists in order: if a position does not match, we go to the next. To
        see an example, see the docstring of the AWS.get_user_identities.

        Example identities (see cloud implementations):
            - AWS: [UserId, AccountId]
            - GCP: [email address + project ID]
            - Azure: [email address + subscription ID]
            - Kubernetes: [context name]

        Example return values:
            - AWS: [[UserId, AccountId]]
            - GCP: [[email address + project ID]]
            - Azure: [[email address + subscription ID]]
            - Kubernetes: [[current active context], [context 2], ...]

        Returns:
            None if the cloud does not have a concept of user identity
            (access protection will be disabled for these clusters);
            otherwise a list of available identities with the current active
            identity being the first element. Most clouds have only one identity
            available, so the returned list will only have one element: the
            current active identity.

        Raises:
            exceptions.CloudUserIdentityError: If the user identity cannot be
                retrieved.
        """
        return None

    @classmethod
    def get_active_user_identity(cls) -> Optional[List[str]]:
        """Returns currently active user identity of this cloud

        See get_user_identities for definition of user identity.

        Returns:
            None if the cloud does not have a concept of user identity;
            otherwise the current active identity.
        """
        identities = cls.get_user_identities()
        return identities[0] if identities is not None else None

    @classmethod
    def get_k8s_credential_name(cls) -> str:
        """Returns the name of the k8s secret storing the credentials for the store."""
        raise NotImplementedError


class CloudStorage:
    """Interface for a cloud object store."""

    # this needs to be overridden by the subclass
    _STORE: typing.Type[AbstractStore]

    def is_directory(self, url: str) -> bool:
        """Returns whether 'url' is a directory.

        In cloud object stores, a "directory" refers to a regular object whose
        name is a prefix of other objects.
        """
        raise NotImplementedError

    def make_sync_dir_command(self, source: str, destination: str) -> str:
        """Makes a runnable bash command to sync a 'directory'."""
        raise NotImplementedError

    def make_sync_file_command(self, source: str, destination: str) -> str:
        """Makes a runnable bash command to sync a file."""
        raise NotImplementedError

    def check_credentials(self):
        """Checks if the user has access credentials to this cloud."""
        return self._STORE.check_credentials()

    def check_credentials_from_secret(self):
        """Checks if the user has access credentials to this cloud."""
        return self._STORE.check_credentials_from_secret()

    def set_secret_credentials(self):
        """Set the credentials from the secret"""
        return self._STORE.set_secret_credentials()



from typing import Any, Dict, List

import socketio
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from kubernetes import client
from kubernetes.client.exceptions import ApiException

from konduktor import logging as konduktor_logging
from konduktor.kube_client import batch_api, core_api, crd_api

from .sockets import socketio as sio

logger = konduktor_logging.get_logger(__name__)

# FastAPI app
app = FastAPI()


# CORS Configuration
app.add_middleware(
    CORSMiddleware,
    allow_origins=['*'],  # Allow all origins
    allow_credentials=True,
    allow_methods=['*'],  # Allow all methods
    allow_headers=['*'],  # Allow all headers
)

# Use Kubernetes API clients
# Initialize BatchV1 and CoreV1 API (native kubernetes)
batch_client = batch_api()
core_client = core_api()
# Initialize Kueue API
crd_client = crd_api()


@app.get('/')
async def home():
    return JSONResponse({'home': '/'})


@app.delete('/deleteJob')
async def delete_job(request: Request):
    data = await request.json()
    name = data.get('name', '')
    namespace = data.get('namespace', 'default')

    try:
        delete_options = client.V1DeleteOptions(propagation_policy='Background')

        crd_client.delete_namespaced_custom_object(
            group='kueue.x-k8s.io',
            version='v1beta1',
            namespace=namespace,
            plural='workloads',
            name=name,
            body=delete_options,
        )
        logger.debug(f"Kueue Workload '{name}' deleted successfully.")

        return JSONResponse({'success': True, 'status': 200})

    except ApiException as e:
        logger.debug(f'Exception: {e}')
        return JSONResponse({'error': str(e)}, status_code=e.status)


@app.get('/getJobs')
async def get_jobs():
    rows = fetch_jobs()
    return JSONResponse(rows)


@app.get('/getNamespaces')
async def get_namespaces():
    try:
        # Get the list of namespaces
        namespaces = core_client.list_namespace()
        # Extract the namespace names from the response
        namespace_list = [ns.metadata.name for ns in namespaces.items]
        return JSONResponse(namespace_list)
    except ApiException as e:
        logger.debug(f'Exception: {e}')
        return JSONResponse({'error': str(e)}, status_code=e.status)


@app.put('/updatePriority')
async def update_priority(request: Request):
    data = await request.json()
    name = data.get('name', '')
    namespace = data.get('namespace', 'default')
    priority = data.get('priority', 0)

    try:
        job = crd_client.get_namespaced_custom_object(
            group='kueue.x-k8s.io',
            version='v1beta1',
            namespace=namespace,
            plural='workloads',
            name=name,
        )

        job['spec']['priority'] = priority

        crd_client.patch_namespaced_custom_object(
            group='kueue.x-k8s.io',
            version='v1beta1',
            namespace=namespace,
            plural='workloads',
            name=name,
            body=job,
        )
        return JSONResponse({'success': True, 'status': 200})

    except ApiException as e:
        logger.debug(f'Exception: {e}')
        return JSONResponse({'error': str(e)}, status_code=e.status)


# Get a listing of workloads in kueue
def fetch_jobs():
    listing = crd_client.list_namespaced_custom_object(
        group='kueue.x-k8s.io',
        version='v1beta1',
        namespace='default',
        plural='workloads',
    )

    return format_workloads(listing)


def format_workloads(listing: Dict[str, Any]) -> List[Dict[str, Any]]:
    if not listing:
        return []

    res = []

    for job in listing['items']:
        id = job['metadata']['uid']
        name = job['metadata']['name']
        created_at = job['metadata']['creationTimestamp']
        namespace = job['metadata']['namespace']
        localQueueName = job['spec'].get('queueName', 'Unknown')
        priority = job['spec']['priority']
        active = job['spec'].get('active', 0)
        status = 'ADMITTED' if 'admission' in job.get('status', {}) else 'PENDING'

        statusVal = 1 if 'admission' in job.get('status', {}) else 0
        order = (statusVal * 10) + priority

        res.append(
            {
                'id': id,
                'name': name,
                'namespace': namespace,
                'localQueueName': localQueueName,
                'priority': priority,
                'status': status,
                'active': active,
                'created_at': created_at,
                'order': order,
            }
        )

    return res


app = socketio.ASGIApp(sio, app)



"""
Utility for (un)zip and encode/decoding k8s secrets in base64
"""

import base64
import os
import shutil
import tempfile
import zipfile
from typing import List


def zip_base64encode(files: List[str]) -> str:
    """Zips files and encodes them in base64.

    Args:
        files: List of file paths to zip. Can include files and directories.

    Returns:
        Base64 encoded string of the zipped files.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        # Copy all files/directories to temp dir preserving structure
        for file_path in files:
            src_path = os.path.expanduser(file_path)
            if not os.path.exists(src_path):
                continue
            dst_path = os.path.join(temp_dir, os.path.basename(file_path))

            if os.path.isdir(src_path):
                shutil.copytree(src_path, dst_path)
            else:
                shutil.copy2(src_path, dst_path)

        # Create zip file
        zip_path = os.path.join(temp_dir, 'archive.zip')
        with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
            for item in os.listdir(temp_dir):
                if item == 'archive.zip':
                    continue
                item_path = os.path.join(temp_dir, item)
                if os.path.isfile(item_path):
                    zipf.write(item_path, item)
                else:
                    for root, _, files in os.walk(item_path):
                        for file in files:
                            file_path = os.path.join(root, file)
                            arcname = os.path.relpath(file_path, temp_dir)
                            zipf.write(file_path, arcname)

        # Read and encode zip file
        with open(zip_path, 'rb') as f:
            zip_str = f.read()
            secret_value = base64.b64encode(zip_str).decode('utf-8')
            # print("encoding")
            # print(type(secret_value))
            # print(len(secret_value))
            # print(secret_value[-20:])
            return secret_value


def base64decode_unzip(secret_value: str, output_path: str) -> str:
    """Decodes a base64 encoded string and unzips the files.

    Args:
        secret_value: Base64 encoded string of the zipped files.
        output_path: Path where to extract the unzipped files.

    Returns:
        Path to the unzipped files.
    """
    # TODO(asaiacai): this is messy I know...
    # Decode base64 string
    # print("decoding")
    # print(type(secret_value))
    # print(len(secret_value))
    # print(secret_value[-20:])
    decoded_data = base64.b64decode(secret_value)

    # Write decoded data to temporary zip file
    with tempfile.TemporaryDirectory() as temp_dir:
        zip_path = os.path.join(temp_dir, 'archive.zip')

        with zipfile.ZipFile(zip_path, 'w') as zipf:
            zipf.writestr('data.zip', decoded_data)

        with zipfile.ZipFile(zip_path, 'r') as zipf:
            zipf.extractall(path=output_path)

        with zipfile.ZipFile(os.path.join(output_path, 'data.zip'), 'r') as zipf:
            zipf.extractall(path=output_path)

    return output_path
